import enum
import os

import numpy as np
from cached_property import cached_property

import homcloud.geometry_resolver as geom_resolver
from homcloud.pdgm_format import PDGMReader, MAGIC_HEADER, MAGIC_HEADER_LENGTH


def empty_pd(degree=0, sign_flipped=False):
    empty_array = np.zeros((0,))
    return SimplePDGM(degree, empty_array, empty_array, empty_array, sign_flipped)


class PDGMBase(object):
    def minmax_of_birthdeath_time(self):
        return (min(self.births.min(), self.deaths.min()), max(self.births.max(), self.deaths.max()))

    def count_pairs_in_rectangle(self, birth1, birth2, death1, death2):
        birth_min = min([birth1, birth2])
        birth_max = max([birth1, birth2])
        death_min = min([death1, death2])
        death_max = max([death1, death2])
        return np.sum(
            (self.births >= birth_min)
            & (self.births <= birth_max)
            & (self.deaths >= death_min)
            & (self.deaths <= death_max)
        )


class SimplePDGM(PDGMBase):
    def __init__(self, degree, births, deaths, ess_births=np.array([]), sign_flipped=False):
        self.degree = degree
        self.births = births
        self.deaths = deaths
        self.essential_births = ess_births
        self.sign_flipped = sign_flipped

    @staticmethod
    def load_from_textfile(path, degree=None, sign_flipped=False):
        with open(path) as f:
            return SimplePDGM.load_from_text(f, degree, sign_flipped)

    @staticmethod
    def load_from_text(infile, degree=None, sign_flipped=False):
        """Create a PD object from text file

        Args:
        infile -- io-like object
        """
        births = []
        deaths = []
        essbirths = []
        for line in infile:
            line = line.strip()
            if not line:
                continue
            if line.startswith("#"):
                continue
            birth, death = line.split()
            if death == "Inf":
                essbirths.append(float(birth))
            elif birth == "-Inf":
                essbirths.append(float(death))
            else:
                births.append(float(birth))
                deaths.append(float(death))

        return SimplePDGM(degree, np.array(births), np.array(deaths), np.array(essbirths), sign_flipped)


class PDGM(PDGMBase):
    def __init__(self, reader, degree, load_indexed_pairs=True):
        self.pdgmreader = reader
        self.degree = degree
        self.load_pd()
        if load_indexed_pairs:
            self.load_indexed_pd()

    def load_pd(self):
        births, deaths, ess_births = self.pdgmreader.load_pd_chunk("pd", self.degree)
        self.births = np.array(births)
        self.deaths = np.array(deaths)
        self.essential_births = np.array(ess_births)

    def load_indexed_pd(self):
        births, deaths, ess_births = self.pdgmreader.load_pd_chunk("indexed_pd", self.degree)
        if births is None:
            return
        self.birth_indices = np.array(births, dtype=int)
        self.death_indices = np.array(deaths, dtype=int)
        self.essential_birth_indices = np.array(ess_births, dtype=int)

    def has_chunk(self, chunktype, degree=None):
        return self.pdgmreader.find_chunk_metadata(chunktype, degree) is not None

    @staticmethod
    def open(path, degree, load_indexed_pairs=True):
        return PDGM(PDGMReader.open(path), degree, load_indexed_pairs)

    def close(self):
        self.pdgmreader.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    @property
    def sign_flipped(self):
        return self.pdgmreader.metadata["sign_flipped"]

    @property
    def filtration_type(self):
        return self.pdgmreader.metadata["filtration_type"]

    @property
    def input_dim(self):
        return self.pdgmreader.metadata["dim"]

    @property
    def num_pairs(self):
        return len(self.births)

    @property
    def death_index_to_pair_number(self):
        return dict(zip(self.death_indices, range(self.num_pairs)))

    @property
    def death_index_to_birth_index(self):
        return dict(zip(self.death_indices, self.birth_indices))

    @cached_property
    def index_to_level(self):
        return self.pdgmreader.load_simple_chunk("index_to_level")

    @cached_property
    def index_to_pixel(self):
        return self.pdgmreader.load_simple_chunk("index_to_pixel")

    @cached_property
    def bitmap_information(self):
        return self.pdgmreader.load_bitmap_information_chunk()

    @cached_property
    def indexed_bitmap(self):
        assert self.filtration_type == "bitmap"
        shape = self.bitmap_information.shape
        bitmap = np.empty(shape, dtype=int)
        for index, pixel in enumerate(self.index_to_pixel):
            bitmap[tuple(pixel)] = index
        return bitmap

    @property
    def index_to_simplex(self):
        return self.pdgmreader.load_simple_chunk("index_to_simplex")

    @cached_property
    def boundary_map_chunk(self):
        return self.pdgmreader.load_boundary_map_chunk()

    def load_boundary_map(self):
        return self.boundary_map_chunk

    @property
    def boundary_map_bytes(self):
        return self.pdgmreader.load_chunk_bytes("boundary_map")

    def default_vertex_symbols(self):
        return [str(k) for k in range(self.num_vertices)]

    @cached_property
    def num_vertices(self):
        if self.filtration_type == "alpha":
            return self.pdgmreader.load_chunk("alpha_information")["num_vertices"]
        if self.filtration_type == "rips":
            return self.pdgmreader.load_chunk("rips_information")["num_vertices"]

    @cached_property
    def simplicial_symbol_resolver(self):
        assert self.filtration_type in ["alpha", "simplicial"]

        return geom_resolver.SimplicialResolver(
            self.pdgmreader.load_simple_chunk("index_to_simplex"),
            self.pdgmreader.load_simple_chunk("vertex_symbols") or self.default_vertex_symbols(),
            self.boundary_map,
        )

    @property
    def alpha_symbol_resolver(self):
        return self.simplicial_symbol_resolver

    @cached_property
    def alpha_coord_resolver(self):
        assert self.filtration_type == "alpha"

        return geom_resolver.SimplicialResolver(
            self.pdgmreader.load_simple_chunk("index_to_simplex"),
            self.pdgmreader.load_simple_chunk("vertex_coordintes"),
            self.boundary_map,
        )

    @cached_property
    def cubical_geometry_resolver(self):
        assert self.filtration_type == "cubical"

        return geom_resolver.CubicalResolver(
            self.pdgmreader.load_chunk("bitmap_information")["shape"],
            self.pdgmreader.load_simple_chunk("index_to_cube"),
            self.boundary_map,
        )

    @cached_property
    def bitmap_geometry_resolver(self):
        assert self.filtration_type == "bitmap"

        return geom_resolver.BitmapResolver(self.index_to_pixel)

    @cached_property
    def abstract_geometry_resolver(self):
        assert self.filtration_type == "abstract"

        return geom_resolver.AbstractResolver(self.pdgmreader.load_simple_chunk("index_to_symbol"), self.boundary_map)

    @cached_property
    def rips_geometry_resolver(self):
        assert self.filtration_type == "rips"

        return geom_resolver.RipsResolver(
            self.pdgmreader.load_simple_chunk("vertex_symbols") or self.default_vertex_symbols()
        )

    def geometry_resolver(self, use_symbol):
        if self.filtration_type == "alpha":
            if use_symbol:
                return self.alpha_symbol_resolver
            else:
                return self.alpha_coord_resolver
        elif self.filtration_type == "simplicial":
            return self.simplicial_symbol_resolver
        elif self.filtration_type == "cubical":
            return self.cubical_geometry_resolver
        elif self.filtration_type == "bitmap":
            return self.bitmap_geometry_resolver
        elif self.filtration_type == "abstract":
            return self.abstract_geometry_resolver
        elif self.filtration_type == "rips":
            return self.rips_geometry_resolver
        else:
            raise RuntimeError("Geometry resolver is unavailable for {}".format(self.filtration_type))

    def boundary_map(self, cell_index):
        return self.boundary_map_chunk["map"][cell_index][1]

    @property
    def path(self):
        return self.pdgmreader.path

    @cached_property
    def birth_positions(self):
        return self.geometry_resolver(False).resolve_cells(self.birth_indices)

    @cached_property
    def death_positions(self):
        return self.geometry_resolver(False).resolve_cells(self.death_indices)

    @cached_property
    def essential_birth_positions(self):
        return self.geometry_resolver(False).resolve_cells(self.essential_birth_indices)

    def pairs_positions(self):
        return zip(self.births, self.deaths, self.birth_positions, self.death_positions)

    @cached_property
    def alpha_information(self):
        return self.pdgmreader.load_chunk("alpha_information")

    @property
    def alpha_weighted(self):
        return self.alpha_information["weighted"]

    @property
    def alpha_radii_squared(self):
        return self.alpha_information["squared"]

    @property
    def graph_adjacent_matrix(self):
        chunk = self.pdgmreader.load_chunk("graph_weights")
        if chunk is None:
            return None
        num_vertices = chunk["num_vertices"]
        matrix = np.zeros((num_vertices, num_vertices), dtype=float)
        xs, ys = np.triu_indices(num_vertices, 1)
        matrix[(xs, ys)] = chunk["weights"]
        matrix[(ys, xs)] = chunk["weights"]
        return matrix

    def representative_cocycle(self, d, nth):
        return self.pdgmreader.load_simple_chunk("cocycles", d)[nth]


class FileType(enum.Enum):
    TEXT = 0
    PDGM = 5
    UNKNOWN = -1

    @staticmethod
    def estimate(path, typestr):
        TABLE_TYPESTR = {
            "text": FileType.TEXT,
            "pdgm": FileType.PDGM,
        }
        TABLE_EXTENSION = {
            ".txt": FileType.TEXT,
            ".pdgm": FileType.PDGM,
        }

        def check_file_header(path):
            with open(path, "rb") as f:
                if f.read(MAGIC_HEADER_LENGTH) == MAGIC_HEADER:
                    return FileType.PDGM
                else:
                    return None

        _, ext = os.path.splitext(path)
        return TABLE_TYPESTR.get(typestr) or check_file_header(path) or TABLE_EXTENSION.get(ext) or PD.FileType.unknown

    def loader(self, degree, sign_flipped):
        if self == FileType.TEXT:
            return lambda path: SimplePDGM.load_from_textfile(path, degree, sign_flipped)
        elif self == FileType.PDGM:
            return lambda path: PDGM.open(path, degree)
        else:
            raise ValueError("Unknown filetype")


def load(path, filetype=None, degree=None, sign_flipped=False):
    filetype = FileType.estimate(path, filetype)
    return filetype.loader(degree, sign_flipped)(path)


def load_merged_diagrams(paths, filetype=None, degree=None, sign_flipped=False):
    def merge_diagrams(diagrams):
        if len(diagrams) == 0:
            return SimplePDGM.empty_pd(degree, sign_flipped)
        births = np.concatenate([pd.births for pd in diagrams])
        deaths = np.concatenate([pd.deaths for pd in diagrams])
        ess_births = np.concatenate([pd.essential_births for pd in diagrams])
        return SimplePDGM(degree, births, deaths, ess_births, sign_flipped)

    filetype = FileType.estimate(paths[0], filetype)
    loader = filetype.loader(degree, sign_flipped)
    diagrams = [loader(path) for path in paths]
    if len(diagrams) == 1:
        return diagrams[0]
    else:
        return merge_diagrams(diagrams)



import argparse
import sys

from homcloud.version import __version__

from homcloud.license import add_argument_for_license
from homcloud.modp_reduction_ext import ModpMatrix
from homcloud.pdgm import PDGM
from homcloud.pdgm_format import PDGMWriter
from homcloud.int_reduction import setup_matrix, count_num_cells


def main(args=None):
    args = args or argument_parser().parse_args()
    pdgm = PDGM.open(args.input, 0, False)
    matrix = build_matrix(args.p, pdgm.input_dim, pdgm.boundary_map_chunk)

    matrix.reduce()
    pairs = matrix.birth_death_pairs()
    with open(args.output, "wb") as f:
        write_modp_pdgm(f, pdgm, pairs)


def argument_parser():
    p = argparse.ArgumentParser(description="Examine field problem")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_argument_for_license(p)
    p.add_argument("-p", type=int, required=True, help="prime number")
    p.add_argument("input", help="input file")
    p.add_argument("output", help="output file")
    return p


def build_matrix(p, dim, boundary_map):
    matrix = ModpMatrix(p, count_num_cells(dim, boundary_map))
    setup_matrix(matrix, boundary_map)
    return matrix


def write_modp_pdgm(f, orig_pdgm, pairs):
    writer = PDGMWriter(f, orig_pdgm.filtration_type, orig_pdgm.input_dim)
    index_to_level = orig_pdgm.index_to_level
    writer.save_pairs(pairs, index_to_level)
    writer.append_simple_chunk("index_to_level", index_to_level)
    writer.write()


if __name__ == "__main__":
    sys.exit(main())



import torch


def sqnorm(a):
    return torch.sum(a**2)


def edge_squared_circumradius(edge):
    return sqnorm(edge[0] - edge[1]) / 4


def triangle_squared_circumradius_2d(triangle):
    a = triangle[0] - triangle[1]
    b = triangle[1] - triangle[2]
    c = triangle[2] - triangle[0]
    cross = a[0] * b[1] - b[0] * a[1]
    return sqnorm(a) * sqnorm(b) * sqnorm(c) / (4 * cross**2)


def triangle_squared_circumradius_3d(triangle):
    a = triangle[0] - triangle[1]
    b = triangle[1] - triangle[2]
    c = triangle[2] - triangle[0]
    cross = torch.cross(a, b)
    return sqnorm(a) * sqnorm(b) * sqnorm(c) / (4 * sqnorm(cross))


def tetrahedron_squared_circumradius_3d(tetrahedron):
    D = torch.vstack(
        [
            torch.sum(tetrahedron**2, 1),
            tetrahedron.transpose(0, 1),
            torch.ones(4, dtype=float),
        ]
    )
    a = torch.det(D[[1, 2, 3, 4]])
    c = torch.det(D[[0, 1, 2, 3]])
    Dx = torch.det(D[[0, 2, 3, 4]])
    Dy = -torch.det(D[[0, 1, 3, 4]])
    Dz = torch.det(D[[0, 1, 2, 4]])
    return (Dx**2 + Dy**2 + Dz**2 - 4 * a * c) / (4 * (a**2))



import argparse

from scipy.sparse import lil_matrix
from scipy.sparse.csgraph import depth_first_order
from cached_property import cached_property
import numpy as np

from homcloud.version import __version__
from homcloud.license import add_argument_for_license
from homcloud.pdgm import PDGM
from homcloud.spatial_searcher import SpatialSearcher


def main(args=None):
    args = args or argument_parser().parse_args()
    with PDGM.open(args.input, 0) as pdgm:
        spatial_searcher = SpatialSearcher(list(zip(pdgm.birth_indices, pdgm.death_indices)), pdgm.births, pdgm.deaths)
        pair = spatial_searcher.nearest_pair(args.x, args.y)
        graph = Graph(pdgm, pair[0], pair[1], args.epsilon)


class Graph(object):
    def __init__(self, pdgm, birth_index, death_index, epsilon):
        self.pdgm = pdgm
        self.boundary_map_chunk = pdgm.boundary_map_chunk
        self.index_to_simplex = pdgm.index_to_simplex
        self.birth_index = birth_index
        self.death_index = death_index
        self.epsilon = epsilon

        assert self.dim(self.birth_index) == 0
        assert self.dim(self.death_index) == 1

        self.build_graph()

    def dim(self, i):
        return self.boundary_map_chunk["map"][i][0]

    def index_to_vertex(self, i):
        return self.index_to_simplex[i][0]

    def edge(self, i):
        x, y = self.boundary_map_chunk["map"][i][1]
        return self.index_to_vertex(x), self.index_to_vertex(y)

    def edge_index_pair(self, i):
        return self.boundary_map_chunk["map"][i][1]

    @property
    def num_vertices(self):
        return self.pdgm.num_vertices

    @property
    def index_to_level(self):
        return self.pdgm.index_to_level

    def build_graph(self):
        death = self.index_to_level[self.death_index]
        self.graph = lil_matrix((self.num_vertices, self.num_vertices), dtype=int)
        for i in range(self.death_index):
            if self.index_to_level[i] > death - self.epsilon:
                break
            if self.dim(i) == 1:
                x, y = self.edge(i)
                self.graph[x, y] = 1

    @cached_property
    def birth_component(self):
        return depth_first_order(self.graph, self.index_to_vertex(self.birth_index), False, False)

    @property
    def elder_component(self):
        x, y = self.edge(self.death_index)
        if np.count_nonzero(self.birth_component == x):
            return depth_first_order(self.graph, y, False, False)
        else:
            return depth_first_order(self.graph, x, False, False)


def argument_parser():
    p = argparse.ArgumentParser(description="Compute maximal volume in PH0")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_argument_for_license(p)
    tp = p.add_argument_group("target")
    tp.add_argument("-x", required=True, type=float, help="birth time of the pair")
    tp.add_argument("-y", required=True, type=float, help="death time of the pair")

    op = p.add_argument_group("output options")
    op.add_argument("-j", "--json-output", help="output in json format")

    cp = p.add_argument_group("computation parameters")
    cp.add_argument("-e", "--epsilon", type=float, default=0.0, help="noise bandwidth (default: 0.0)")

    p.add_argument("input", help="input filename")
    return p


if __name__ == "__main__":
    main()



import argparse
import json

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import numpy as np
from homcloud.plot_PD import add_arguments_for_zspec, add_arguments_for_auxinfo, ZSpec, AuxPlotInfo, PDPlotter
from homcloud.version import __version__
from homcloud.histogram import Histogram


def main(args=None):
    args = args or argument_parser().parse_args()
    vector = np.loadtxt(args.input, ndmin=1)
    args.diffuse_pairs = True
    with open(args.vectorization_info) as f:
        info = json.load(f)

    zspec = ZSpec.create_from_args(args)
    histogram = Histogram.reconstruct_from_vector(vector, info)
    aux_info = AuxPlotInfo.from_args(args)
    plotter = PDPlotter.find_plotter(args.style)(histogram, zspec, aux_info)

    plotter.plot(*plt.subplots())

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.show()


def argument_parser():
    p = argparse.ArgumentParser(description="View vectorized histogram")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument("-o", "--output", help="output image file")
    add_arguments_for_zspec(p)
    p.add_argument("-s", "--style", default="colorhistogram", help="plotting style (colorhistogram(default), contour)")
    add_arguments_for_auxinfo(p)
    p.add_argument(
        "--dpi", type=int, default=None, help="output DPI (used with -o option, default is savefig.dpi for matplotlib)"
    )
    p.add_argument("input", help="Vector data file")
    p.add_argument("vectorization_info", help="vectorization information json data")
    return p


def draw_diagonal(ax, x_edges):
    x_range = (x_edges[0], x_edges[-1])
    ax.add_line(Line2D(x_range, x_range, linewidth=1, color="black"))


if __name__ == "__main__":
    main()



import numpy as np

import homcloud.pict.tree as pict_tree
import homcloud.paraview_interface as pv_interface
import homcloud.plotly_3d as p3d
from homcloud.spatial_searcher import SpatialSearcher
from . import bitmap as bm


class BitmapPHTrees(object):
    """
    This class represents PH trees computed from a bitmap filtration.

    You can create an instance of this class by :meth:`PDList.bitmap_phtrees`.


    Attributes:
        nodes (list of :class:`BitmapPHTrees.Node`): All nodes
        essential_nodes (list of :class:`BitmapPHTrees.Node`): All essential nodes
        nonessential_nodes (list of :class:`BitmapPHTrees.Node`): All nonessential nodes
        roots (list of :class:`BitmapPHTrees.Node`): All root nodes
    """

    @staticmethod
    def for_bitmap_levelset(array, mode="sublevel", save_to=None):
        """Computes a PH trees from 0-th and (n-1)-st persistent homology
        from levelset filtration of the bitmap.

        Args:
            array (numpy.ndarray): The bitmap data.
            mode (string): "superlevel" or "sublevel".
            save_to (string or None): The filepath which the
                PH trees is stored in.

        Returns:
            :class:`PDList`: The 0th and (n-1)st PDs with PH-trees.
        """
        from .pd import PDList

        is_superlevel = mode == "superlevel"
        lower, upper = pict_tree.construct_mergetrees(array, is_superlevel)

        f = PDList.open_pdgm_file(save_to)
        pict_tree.save_pdgm(f, array.ndim, is_superlevel, lower, upper)
        f.seek(0)
        return PDList(f, PDList.FileType.PDGM)

    def __init__(self, treedict, is_superlevel):
        self.treedict = treedict
        self.is_superlevel = is_superlevel

        self.nodes = [self.node(id) for id in self.treedict["nodes"] if not self.is_trivial_id(id)]
        self.essential_nodes = [node for node in self.nodes if node.essential()]
        self.nonessential_nodes = [node for node in self.nodes if not node.essential()]
        self.spatial_searcher = self.build_spatial_searcher()
        self.roots = [node for node in self.nodes if node.isroot()]

    def node(self, id):
        return BitmapPHTrees.Node(self, self.treedict["nodes"][id])

    def is_trivial_id(self, id):
        node = self.treedict["nodes"][id]
        return node["birth-time"] == node["death-time"]

    def build_spatial_searcher(self):
        births = np.array([node.birth_time() for node in self.nonessential_nodes])
        deaths = np.array([node.death_time() for node in self.nonessential_nodes])
        return SpatialSearcher(self.nonessential_nodes, births, deaths)

    def degree(self):
        """
        Returns:
            int: The degree of the persistent homology.
        """
        return self.treedict["degree"]

    def nearest_pair_node(self, x, y):
        """
        Searches a tree node corresponding the birth-death pair
        nearest to (x, y).

        Args:
            x (float): The x-coordinate.
            y (float): The y-coordinate.
        Returns:
            :class:`BitmapPHTrees.Node`: The node.
        """
        return self.spatial_searcher.nearest_pair(x, y)

    def pair_nodes_in_rectangle(self, xmin, xmax, ymin, ymax):
        """
        Searches tree nodes corresponding the birth-death pairs
        which the given rectangle contains.

        Args:
            xmin (float): The left side of the rectangle.
            xmax (float): The right side of the rectangle.
            ymin (float): The bottom side of the rectangle.
            ymax (float): The top side of the rectangle.

        Returns:
            list of :class:`BitmapPHTrees.Node`: The nodes.
        """

        return self.spatial_searcher.in_rectangle(xmin, xmax, ymin, ymax)

    def sign(self):
        return -1 if self.is_superlevel else 1

    class Volume(object):
        """
        This class is the superclass of :class:`Node` and :class:`StableVolume`.

        This class defines common methods for these two classes.
        """

        def to_paraview_node(self, gui_name=None):
            """
            Construct a :class:`homcloud.paraview_interface.PipelineNode` object
            to visulize :meth:`volume`.

            You can show the volume by
            :meth:`homcloud.paraview_interface.show`. You can also
            adjust the visual by the methods of
            :class:`homcloud.paraview_interface.PipelineNode`.

            Args:
                gui_name (string or None): The name shown in Pipeline Browser
                    in paraview's GUI.

            Returns:
                homcloud.paraview_interface.PipelineNode: A PipelineNode object.
            """
            return BitmapPHTrees.to_paraview_node_from_nodes([self], gui_name)

        to_pvnode = to_paraview_node

        def to_plotly3d_trace(self, color=None, name=""):
            """
            Constructs a plotly's trace object to visualize the volume

            Args:
                color (string or None): The name of the color
                name (string): The name of the object

            Returns:
                plotly.graph_objects.Mesh3d: Plotly's trace object
            """
            return p3d.Voxels(self.volume(), color, name)

        to_plotly3d = to_plotly3d_trace

        def to_pyvista_mesh(self):
            """
            Constructs a PyVista's mesh object to visualize an optimal/stable volume

            Returns:
                pyvista.PolyData: PyVista's mesh object
            """
            import homcloud.pyvistahelper as pvhelper

            return pvhelper.SparseVoxels(self.volume())

    class Node(Volume):
        """
        This class represents a tree node of PH trees for a bitmap filtration.

        You can draw the optimal volume corresponding to the node
        on an image by :meth:`draw_volumes_on_2d_image`.
        """

        def __init__(self, mt, dic):
            self.mt = mt
            self.dic = dic

        def birth_time(self):
            """
            Returns:
                float: The birth time.
            """
            return self.dic["birth-time"]

        def death_time(self):
            """
            Returns:
                float: The death time. May be infinity.
            """
            if not self.essential():
                return self.dic["death-time"]
            return self.mt.sign() * np.inf

        def lifetime(self):
            """
            Returns:
                float: The lifetime of the pair.
            """
            return self.death_time() - self.birth_time()

        def __iter__(self):
            return (self.birth_time(), self.death_time()).__iter__()

        def birth_position(self):
            """Returns the birth pixel.

            Returns:
               list of int: The coordinate of the birth pixel.
            """
            return self.dic["birth-pixel"]

        birth_pixel = birth_position

        def death_position(self):
            """Returns the death pixel.

            Returns:
               list of int: The coordinate of the death pixel.
            """
            return self.dic["death-pixel"]

        death_pixel = death_position

        def volume(self):
            """
            Returns the optimal volume.

            In fact, for degree 0 node, this method returns
            optimal 0-cohomological volume (this means
            maximal connected conponents in the filtration)
            and for degree (n-1),
            this method returns optimal (n-1)-homological volume.
            This fact is quite mathematical problem, so if you feel that
            it is too difficult to understand, you can ignore the fact.
            You can understand the volume information without the understanding
            of such mathematical background.

            Returns:
                list of list of int: The coordinates of all pixels
                in the optimal volume.
            """
            return self.dic["volume"]

        def stable_volume(self, epsilon):
            """
            Returns the stable volume of the optimal volume.

            Args:
                epsilon (float): Duration noise strength

            Returns:
                :class:`BitmapPHTrees.StableVolume`: The stable volume

            """
            return BitmapPHTrees.StableVolume(self, epsilon)

        def essential(self):
            """
            Returns:
                bool: True if the death time is infinity.
            """
            return self.dic["death-time"] is None

        def parent(self):
            """
            Returns:
                BitmapPHTrees.Node: The parent node of the node in the PH tree.
            """
            parent_id = self.dic["parent"]
            return BitmapPHTrees.Node(self.mt, self.mt.treedict["nodes"][parent_id])

        def isroot(self):
            """
            Returns:
               bool: True if the node is a root node.
            """
            return self.dic["parent"] is None

        def children(self):
            """
            Returns:
                list of :class:`BitmapPHTrees.Node`: All children nodes.
            """
            return [self.mt.node(child_id) for child_id in self.dic["children"] if not self.mt.is_trivial_id(child_id)]

        def __eq__(self, other):
            return (self.mt == other.mt) and (self.dic["id"] == other.dic["id"])

        def __repr__(self):
            return "BitmapPHTrees.Node({}, {})".format(self.birth_time(), self.death_time())

    class StableVolume(Volume):
        """
        This class represents a stable volume in :class:`BitmapPHTrees`.
        """

        def __init__(self, node, epsilon):
            self.node = node

            self._volume = set(tuple(p) for p in node.dic["volume"])
            for child_id in node.dic["children"]:
                child_dic = node.mt.treedict["nodes"][child_id]
                if self.is_unstable_child(child_dic, epsilon):
                    self._volume.difference_update(tuple(p) for p in child_dic["volume"])

        def is_unstable_child(self, child_dic, epsilon):
            is_sublevel = not self.node.mt.is_superlevel
            is_lower = self.node.mt.degree() == 0
            if is_lower and is_sublevel:
                return child_dic["death-time"] > self.node.death_time() - epsilon
            elif is_lower and not is_sublevel:
                return child_dic["death-time"] < self.node.death_time() + epsilon
            elif not is_lower and is_sublevel:
                return child_dic["birth-time"] < self.node.birth_time() + epsilon
            elif not is_lower and not is_sublevel:
                return child_dic["birth-time"] > self.node.birth_time() - epsilon

        def volume(self):
            """
            Returns the pixels of the stable volume.

            Returns:
                list[tuple[int,...]]: The coordinates of all pixels.
            """
            return list(self._volume)

    @staticmethod
    def to_paraview_node_from_nodes(nodes, gui_name=None):
        """
        Construct a :class:`homcloud.paraview_interface.PipelineNode`
        object to visulize node volumes of the nodes.

        Args:
            nodes (list of :class:`BitmapPHTrees.Node`):
                The list of nodes to be visualized.
            gui_name (string or None): The name shown in Pipeline Browser
                in paraview's GUI.

        Returns:
            homcloud.paraview_interface.VTK: Paraview Pipeline node object.
        """
        all_volume = []
        for node in nodes:
            all_volume.extend(node.volume())

        min, max = bm.volume_range(all_volume)
        bitmap = bm.volume_to_bitmap(all_volume, (min, max))

        return pv_interface.VoxelData(bitmap, gui_name, min).threshold("value", (0.5, 1.0))

    to_pvnode_from_nodes = to_paraview_node_from_nodes



import argparse

from homcloud.pict import utils
from homcloud.version import __version__
from homcloud.license import add_argument_for_license
from homcloud.argparse_common import (
    parse_cubical_periodic_flags,
    parse_bool,
    check_abolished_output,
    check_cubical_option,
)
from homcloud.bitmap import Bitmap


def main(args=None):
    args = args or argument_parser().parse_args()
    check_abolished_output(args)
    check_cubical_option(args)

    pict = utils.load_nd_picture(args.input, args.type).astype(float)
    if args.mode == "superlevel":
        pict = -pict
        flip_sign = True
    elif args.mode == "sublevel":
        flip_sign = False
    else:
        raise RuntimeError("mode must be superlevel or subelvel")

    bitmap = Bitmap(pict, flip_sign, args.periodic, args.save_boundary_map)
    filt = bitmap.build_filtration(args.cubical)

    filt.compute_pdgm_and_save(args.output, args.algorithm)


def argument_parser():
    p = argparse.ArgumentParser()
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument(
        "-m", "--mode", default="sublevel", help="the way to construct dipha complex (sublevel(default),superlevel)"
    )
    p.add_argument(
        "-I",
        "--combine-index-map",
        default=False,
        action="store_true",
        help="combine the index map with the output filtration",
    )
    p.add_argument(
        "-T", "--type", default="text_nd", help="input data format (text2d,text_nd(default),picture2d,pictures3d,npy)"
    )
    p.add_argument(
        "-D",
        "--convert-to-diagram",
        default=False,
        action="store_true",
        help="call dipha and directly convert to a diagram",
    )
    p.add_argument(
        "-C", "--cubical", default=False, action="store_true", help="Use explicit cubical filtration (slow)"
    )
    p.add_argument("-o", "--output", required=True, help="output file")
    p.add_argument(
        "-p",
        "--periodic",
        default=None,
        type=parse_cubical_periodic_flags,
        help="periodic (example: 0,1,1 for y and z are periodic)",
    )
    p.add_argument("--algorithm", default=None, help="algorithm (dipha, phat-twist, phat-chunk-parallel)")
    p.add_argument(
        "-M",
        "--save-boundary-map",
        default=False,
        type=parse_bool,
        help="save boundary map into idiagram file" "(only available with phat-* algorithms, on/*off*)",
    )
    add_argument_for_license(p)
    p.add_argument("input", nargs="*", help="input files")
    return p


if __name__ == "__main__":
    main()



import numpy as np

import homcloud.plotly_3d as p3d


def circumsphere(tetrahedron):
    """Compute circumsphere

    Args:
        tetrahedron (list of list of double): list of coordinates of vertices

    Returns:
        tuple[np.array, float]: The cneter and radius of the circumsphere

    Example:
    >>> circumsphere([
    >>>     [0, 0, 0],
    >>>     [2, 0, 0],
    >>>     [1, np.sqrt(3), 0],
    >>>     [1, np.sqrt(3) / 3, np.sqrt(6) * 2 / 4.5],
    >>> ])
    -> (array([ 1.        ,  0.57735027, -0.06804138]), 1.156703489647612)
    """
    tetrahedron = np.array(tetrahedron)
    x = tetrahedron[:, 0]
    y = tetrahedron[:, 1]
    z = tetrahedron[:, 2]
    a = np.linalg.det(np.vstack([x, y, z, np.ones(4)]))
    w = x**2 + y**2 + z**2
    Dx = np.linalg.det(np.vstack([w, y, z, np.ones(4)]))
    Dy = -np.linalg.det(np.vstack([w, x, z, np.ones(4)]))
    Dz = np.linalg.det(np.vstack([w, x, y, np.ones(4)]))
    c = np.linalg.det(np.vstack([w, x, y, z]))
    r = np.sqrt(np.inner(Dx, Dx) + np.inner(Dy, Dy) + np.inner(Dz, Dz) - 4 * a * c) / (2 * np.abs(a))
    center = np.array([Dx, Dy, Dz]) / (2 * a)
    np.testing.assert_allclose(np.linalg.norm(center - tetrahedron[0, :]), r)
    np.testing.assert_allclose(np.linalg.norm(center - tetrahedron[1, :]), r)
    np.testing.assert_allclose(np.linalg.norm(center - tetrahedron[2, :]), r)
    np.testing.assert_allclose(np.linalg.norm(center - tetrahedron[3, :]), r)
    return center, r


def circumcircle(tetrahedron, i, j, k):
    """Computes the circumcircle of the triangle in the given tetrahedron
    Args:
        tetrahedron (list of list of double): list of coordinates of vertices
        i (int): The first vertex index, from 0 to 3
        j (int): The second vertex index, from 0 to 3
        k (int): The third vertex index, from 0 to 3

    Returns:
        tuple[np.array, float]: The cneter and radius of the circumcircle

    """
    a = np.array(tetrahedron[i])
    b = np.array(tetrahedron[j])
    c = np.array(tetrahedron[k])
    ac = c - a
    ab = b - a
    g = np.cross(ab, ac)
    d = (np.cross(g, ab) * (np.inner(ac, ac)) + np.cross(ac, g) * (np.inner(ab, ab))) / (2 * np.inner(g, g))
    np.testing.assert_allclose(np.linalg.norm(a + d - b), np.linalg.norm(d))
    np.testing.assert_allclose(np.linalg.norm(a + d - c), np.linalg.norm(d))
    return a + d, np.linalg.norm(d)


def center(tetrahedron, i, j):
    """Computes the center of the edge in the given tetrahedron
    Args:
        tetrahedron (list of list of double): list of coordinates of vertices
        i (int): The first vertex index, from 0 to 3
        j (int): The second vertex index, from 0 to 3

    Returns:
        tuple[np.array, float]: The cneter and radius


    """

    a = np.array(tetrahedron[i])
    b = np.array(tetrahedron[j])
    return (a + b) / 2, np.linalg.norm(b - a) / 2


def minimum_enclosing_ball(tetrahedron):
    """Compute minimum enclosing ball of a tetrahedron

    Args:
        tetrahedron (list of list of double): list of coordinates of vertices

    Returns:
        tuple[np.array, float]: The cneter and radius of the minimum enclosing ball

    Example:
    >>> circumsphere([
    >>>     [0, 0, 0],
    >>>     [2, 0, 0],
    >>>     [1, np.sqrt(3), 0],
    >>>     [1, np.sqrt(3) / 3, np.sqrt(6) * 2 / 4.5],
    >>> ])
    -> (array([ 1.        ,  0.57735027, -0.06804138]), 1.156703489647612)
    """

    def contain(c, r):
        r *= 1.000000001
        return (
            (np.linalg.norm(c - tetrahedron[0]) <= r)
            and (np.linalg.norm(c - tetrahedron[1]) <= r)
            and (np.linalg.norm(c - tetrahedron[2]) <= r)
            and (np.linalg.norm(c - tetrahedron[3]) <= r)
        )

    def find(candidates):
        return min((c for c in candidates if contain(*c)), default=None, key=lambda cand: cand[1])

    candidate = find(center(tetrahedron, i, j) for i in range(4) for j in range(i + 1, 4))
    if candidate is not None:
        return candidate

    candidate = find(
        [
            circumcircle(tetrahedron, 0, 1, 2),
            circumcircle(tetrahedron, 0, 1, 3),
            circumcircle(tetrahedron, 0, 2, 3),
            circumcircle(tetrahedron, 1, 2, 3),
        ]
    )
    if candidate is not None:
        return candidate

    return circumsphere(tetrahedron)


def show_circumcenters(tetrahedron):
    """Method for debug"""
    c0, _ = circumsphere(tetrahedron)
    c1, _ = circumcircle(tetrahedron, 0, 1, 2)
    c2, _ = circumcircle(tetrahedron, 0, 1, 3)
    c3, _ = circumcircle(tetrahedron, 0, 2, 3)
    c4, _ = circumcircle(tetrahedron, 1, 2, 3)
    centers = [center(tetrahedron, i, j)[0] for i in range(4) for j in range(i + 1, 4)]
    cm, _ = minimum_enclosing_ball(tetrahedron)
    fig = go.Figure(
        [
            p3d.Simplices([tetrahedron], color="green", name="tetrahedron"),
            p3d.PointCloud(np.array([c0]), color="black", size=2),
            p3d.PointCloud(np.vstack([c1, c2, c3, c4]), color="red", size=2),
            p3d.PointCloud(np.vstack(centers), color="green", size=2),
            p3d.PointCloud(np.array([cm]), color="blue", size=3),
        ]
    )
    return fig



import re
from functools import reduce
import operator

import numpy as np
from homcloud.compat import imageio


def load_nd_picture(paths, inputtype):
    def check_single_file():
        if len(paths) != 1:
            raise RuntimeError("text2d/text_nd/picture2d/npy require only a single file")

    if inputtype == "pictures3d":
        return stack_picts([load_picture(path) for path in paths])
    if inputtype == "picture2d":
        check_single_file()
        return load_picture(paths[0])
    if inputtype == "text2d":
        check_single_file()
        return np.loadtxt(paths[0])
    if inputtype == "text_nd":
        check_single_file()
        return load_nd_picture_from_text(paths[0])
    if inputtype == "npy":
        check_single_file()
        return np.load(paths[0])

    raise RuntimeError("Unknown file type: {}".format(inputtype))


def load_nd_picture_from_text(path):
    with open(path) as f:
        return read_nd_picture_text(f)


def read_nd_picture_text(f):
    while True:
        line = f.readline()
        if line is None:
            raise ValueError("Format error")

        if re.match(r"\s+#", line) or re.match(r"\s+\Z", line):
            continue

        shape = np.fromstring(line, sep=" ", dtype=int)[::-1]
        return np.loadtxt(f).reshape(shape)


def stack_picts(picts):
    """Stacking pictures into one ndarray"""
    if not all([pict.shape == picts[0].shape for pict in picts]):
        raise RuntimeError("Sizes of all pictures should be same")
    return np.array(picts)


def load_picture(path, filetype="picture"):
    """Load a picture and return 2D ndarray whose values are less than 256."""
    if filetype == "picture":
        pict = imageio.imread(path, as_gray=True)
        if np.max(pict) > 255:
            pict //= 256
        return pict
    elif filetype == "text":
        return np.loadtxt(path)


def nd_indices(shape):
    """Create an array of all indices of the array of given shape.

    Example:
    nd_indices((2, 3)
    # => np.array([[0, 0], [0, 1], [0, 2],
    #              [1, 0], [1, 1], [1, 2]])
    """
    n = reduce(operator.mul, shape, 1)
    return np.indices(shape).reshape(len(shape), n).transpose()


def build_levelset_filtration(pict, cubical, periodic, superlevel, save_boundary_map):
    from homcloud.bitmap import Bitmap

    if cubical or periodic or save_boundary_map:
        return Bitmap(pict, superlevel, periodic, save_boundary_map).build_cubical_filtration()
    else:
        return Bitmap(pict, superlevel).build_bitmap_filtration()



import sys
from argparse import ArgumentParser
import base64

import numpy as np

from homcloud.version import __version__
import homcloud.pict.utils as utils
from homcloud.license import add_argument_for_license


def main(args=None):
    args = args or argument_parser().parse_args()
    pict = utils.load_nd_picture(args.input, args.type)
    if pict.ndim != 3:
        raise RuntimeError("Input picture should be 3d")

    with open(args.output, "w") as f:
        write_vtk_file(f, args.input[0], pict)


def write_vtk_file(f, path, pict):
    zw, yw, xw = pict.shape

    def at(x, y, z):
        return x * (zw + 1) * (yw + 1) + y * (zw + 1) + z

    f.write("# vtk DataFile Version 3.0\n")
    f.write("%s\n" % path)
    f.write("ASCII\n")
    f.write("DATASET UNSTRUCTURED_GRID\n")
    f.write("\n")

    f.write("POINTS %d float\n" % ((zw + 1) * (yw + 1) * (xw + 1)))
    for x in range(xw + 1):
        for y in range(yw + 1):
            for z in range(zw + 1):
                f.write("%f %f %f\n" % (x - 0.5, y - 0.5, z - 0.5))
    f.write("\n")
    f.write("CELLS %d %d\n" % (xw * yw * zw, xw * yw * zw * 9))
    for x in range(xw):
        for y in range(yw):
            for z in range(zw):
                f.write(
                    "8 %d %d %d %d %d %d %d %d\n"
                    % (
                        at(x, y, z),
                        at(x + 1, y, z),
                        at(x, y + 1, z),
                        at(x + 1, y + 1, z),
                        at(x, y, z + 1),
                        at(x + 1, y, z + 1),
                        at(x, y + 1, z + 1),
                        at(x + 1, y + 1, z + 1),
                    )
                )
    f.write("\n")
    f.write("CELL_TYPES %d\n" % (xw * yw * zw))
    for _ in range(xw * yw * zw):
        f.write("11\n")
    f.write("\n")
    f.write("CELL_DATA %d" % (xw * yw * zw))
    f.write("SCALARS value float 1\n")
    f.write("LOOKUP_TABLE default\n")
    for value in pict.ravel():
        f.write("%f\n" % value)


def extend_2d_3d(array, default_value):
    if len(array) == 2:
        return default_value, array[0], array[1]
    else:
        return array


def write_vti_xmlfile(f, pict, offsets=[0.0, 0.0, 0.0]):
    zw, yw, xw = extend_2d_3d(pict.shape, 1)
    offsets = extend_2d_3d(offsets, 0.0)

    header = """<?xml version="1.0"?>
    <VTKFile type="ImageData" byte_order="LittleEndian" header_type="UInt64">
      <ImageData WholeExtent="0 {0} 0 {1} 0 {2}"
                 Origin="{3} {4} {5}"
                 Spacing="1.0 1.0 1.0">
        <Piece Extent="0 {0} 0 {1} 0 {2}">
          <CellData Scalars="value">
            <DataArray Name="value" type="Float64" format="ascii">
    """.format(
        xw, yw, zw, offsets[2] - 0.5, offsets[1] - 0.5, offsets[0] - 0.5
    )

    footer = """
            </DataArray>
          </CellData>
        </Piece>
      </ImageData>
    </VTKFile>
    """
    f.write(header)
    np.savetxt(f, pict.astype(np.float64).ravel())
    f.write(footer)


def argument_parser():
    p = ArgumentParser(description="Convert 3d picture to a vtk file")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument(
        "-T",
        "--type",
        default="text_nd",
        help="input data format " "(text2d,text_nd(default),picture2d,pictures3d,npy)",
    )
    p.add_argument("-o", "--output", required=True, help="output file name")
    add_argument_for_license(p)
    p.add_argument("input", nargs="*", help="input files")
    return p


if __name__ == "__main__":
    sys.exit(main())



class GraphOptimal1Cycle(object):
    """The class represents an optimal (not volume-optimal) 1-cycle computed from graph information.

    You can aquaire an optimal one cycle by :meth:`Pair.optimal_1_cycle`.

    Examples:
        >>> import numpy as np
        >>> import homcloud.interface as hc
        >>> distance_matrix = np.loadtxt("distance_matrix.txt")
        >>> pdlist = hc.PDList.from_rips_filtration(distance_matrix, maxdim=1, save_graph=True)
        >>> pd1 = pdlist.dth_diagram(1)
        >>> pair = pd1.nearest_pair_to(1.2, 1.5)
        >>> optimal_1_cycle = pair.optimal_1_cycle()
        >>> optimal_1_cycle.boundary_points()
        => [0, 2, 41, 17, 9]
    """

    def __init__(self, pair, path_vertices):
        self.pair = pair
        self.path_vertices = path_vertices

    def birth_time(self):
        """Returns the birth time of the pair.

        Returns:
            float: The birth time
        """
        return self.pair.birth_time()

    def death_time(self):
        """Returns the death time of the pair.

        Returns:
            float: The death time
        """
        return self.pair.death_time()

    def boundary_points(self):
        """Returns the points on the 1-cycle.

        Returns:
            list of int: The list of points on the cycle.
            Each point is represented by a number, which is assigned in 0-origin.
        """
        return self.path_vertices

    def boundary_points_symbols(self):
        return self.geometry_resolver(False).resolve_graph_path(self.path_vertices)

    def geometry_resolver(self, use_symbol):
        return self.pair.diagram.pd.geometry_resolver(use_symbol)



import argparse
from homcloud.pdgm_format import PDGMReader
from pprint import pprint


def main(args=None):
    if not args:
        args = argument_parser().parse_args()

    with PDGMReader.open(args.input) as reader:
        pprint(reader.metadata)


def argument_parser():
    p = argparse.ArgumentParser(description="show pdgm metadata for debug")
    p.add_argument("input", help="input file path")
    return p


if __name__ == "__main__":
    main()



import argparse
import json
import os
from tempfile import TemporaryDirectory

from cached_property import cached_property
import numpy as np

from homcloud.version import __version__
from homcloud.argparse_common import parse_range
from homcloud.pdgm_format import PDGMReader
from homcloud.pdgm import PDGM
from homcloud.spatial_searcher import SpatialSearcher
import homcloud.utils as utils


class Volume(object):
    def birth_simplex(self, by="coordinates"):
        return self.geometry_resolver(by).resolve_cell(self.birth_index)

    def death_simplex(self, by="coordinates"):
        return self.geometry_resolver(by).resolve_cell(self.death_index)

    def boundary(self, by="coordinates"):
        return self.geometry_resolver(by).boundary(self.volume_indices())

    def boundary_vertices(self, by="coordinates"):
        return self.geometry_resolver(by).boundary_vertices(self.volume_indices())

    def vertices(self, by="coordinates"):
        return self.geometry_resolver(by).resolve_vertices(self.volume_indices())

    def simplices(self, by="coordinates"):
        return self.geometry_resolver(by).resolve_cells(self.volume_indices())

    def volume_indices(self):
        return (n.death_index for n in self.volume_nodes)

    def to_dict(self):
        return {
            "birth-index": self.birth_index,
            "death-index": self.death_index,
            "birth-time": self.birth_time(),
            "death-time": self.death_time(),
            "boundary": self.boundary(),
            "boundary-by-symbols": self.boundary("symbols"),
            "boundary-vertices": self.boundary_vertices(),
            "boundary-vertices-by-symbols": self.boundary_vertices("symbols"),
            "vertices": self.vertices(),
            "vertices-by-symbols": self.vertices("symbols"),
            "simplices": self.simplices(),
            "simplices-by-symbols": self.simplices("symbols"),
            "children": self.children_dicts(),
        }

    def to_child_dict(self):
        return {
            "birth-index": self.birth_index,
            "death-index": self.death_index,
            "birth-time": self.birth_time(),
            "death-time": self.death_time(),
            "children": self.children_dicts(),
        }

    def children_dicts(self):
        return [child.to_child_dict() for child in self.children]

    def draw_volume(self, drawer, color, **thvalues):
        self.geometry_resolver().draw_boundary(drawer, self.volume_indices(), color, **thvalues)

    def __repr__(self):
        return "{}.{}(birth_time={}, death_time={})".format(
            self.__class__.__module__, self.__class__.__name__, self.birth_time(), self.death_time()
        )


class Node(Volume):
    def __init__(self, birth_index, death_index, parent_death, trees=None):
        self.birth_index = birth_index
        self.death_index = death_index
        self.parent_death = parent_death
        self.trees = trees
        self.volume_cache = None
        self.children = []

    def isroot(self):
        return self.parent_death == np.inf

    def birth_time(self):
        return self.trees.index_to_level[self.birth_index]

    def death_time(self):
        return self.trees.index_to_level[self.death_index]

    def lifetime(self):
        return self.death_time() - self.birth_time()

    @cached_property
    def volume_nodes(self):
        volume_nodes = []

        def iter(n):
            volume_nodes.append(n)
            for child in n.children:
                iter(child)

        iter(self)
        return volume_nodes

    def volume_size(self):
        return len(self.volume_nodes)

    def stable_volume(self, epsilon, cls=None):
        cls = cls or StableVolume
        return cls(self, [child for child in self.children if child.birth_time() > self.birth_time() + epsilon])

    def geometry_resolver(self, by="coordinates"):
        return self.trees.geometry_resolver(by)


class StableVolume(Volume):
    def __init__(self, root, children):
        self.root = root
        self.children = children
        self.volume_cache = None

    @property
    def trees(self):
        return self.root.trees

    @property
    def birth_index(self):
        return self.root.birth_index

    @property
    def death_index(self):
        return self.root.death_index

    def birth_time(self):
        return self.root.birth_time()

    def death_time(self):
        return self.root.death_time()

    def lifetime(self):
        return self.death_time() - self.birth_time()

    def geometry_resolver(self, by="coordinates"):
        return self.root.geometry_resolver(by)

    @cached_property
    def volume_nodes(self):
        volume_nodes = [self.root]
        for child in self.children:
            volume_nodes.extend(child.volume_nodes)
        return volume_nodes


class PHTrees(object):
    def __init__(self, nodes, index_to_level=None, coord_resolver=None, symbol_resolver=None, nodeclass=Node):
        assert nodes is not None
        self.nodes = {
            death_index: nodeclass(birth_index, death_index, parent_death, self)
            for (birth_index, death_index, parent_death) in nodes
        }
        self.index_to_level = index_to_level
        self.coord_resolver = coord_resolver
        self.symbol_resolver = symbol_resolver
        self.roots = []
        self.build_tree()

    @staticmethod
    def from_pdgmreader(reader):
        return PHTrees(
            reader.load_simple_chunk("phtrees"),
            reader.load_simple_chunk("index_to_level"),
        )

    @staticmethod
    def from_pdgm(pdgm, nodeclass=Node):
        return PHTrees(
            pdgm.pdgmreader.load_simple_chunk("phtrees"),
            pdgm.index_to_level,
            pdgm.alpha_coord_resolver,
            pdgm.alpha_symbol_resolver,
            nodeclass,
        )

    def build_tree(self):
        for node in self.nodes.values():
            if node.isroot():
                self.roots.append(node)
            else:
                self.nodes[node.parent_death].children.append(node)

    def parent_of(self, node):
        if node.isroot():
            return None
        return self.nodes[node.parent_death]

    def geometry_resolver(self, type):
        if type == "coordinates":
            return self.coord_resolver
        if type == "symbols":
            return self.symbol_resolver
        raise ValueError("Unknown type: {}".format(type))


class Query(object):
    def __init__(self, query, volume_getter, searcher, input_dim=None, query_ancestors=False, query_children=False):
        self.query = query
        self.volume_getter = volume_getter
        self.searcher = searcher
        self.input_dim = input_dim
        self.query_ancestors = query_ancestors
        self.query_children = query_children
        self.result = []

    @property
    def degree(self):
        return self.phtrees.degree

    def to_dict(self):
        return {
            "format-version": 2,
            "query": {
                "query-type": "signle",
                "query-target": self.volume_getter.query_target_name,
                "degree": self.input_dim - 1,
                "birth": self.query[0],
                "death": self.query[1],
                "ancestor-pairs": self.query_ancestors,
                "query-children": self.query_children,
            },
            "dimension": self.input_dim,
            "result": [r.to_dict() for r in self.result],
        }


class PointQuery(Query):
    def invoke(self):
        death_index = self.searcher.nearest_pair(*self.query)
        self.result = [self.volume_getter(death_index)]


class RectangleQuery(Query):
    def invoke(self):
        death_indices = self.searcher.in_rectangle(
            self.query[0][0], self.query[0][1], self.query[1][0], self.query[1][1]
        )
        self.result = [self.volume_getter(i) for i in death_indices]


class GetOptimalVolume(object):
    def __init__(self, phtrees):
        self.phtrees = phtrees

    def __call__(self, death_index):
        return self.phtrees.nodes[death_index]

    query_target_name = "optimal-volume"


class GetStableVolume(object):
    def __init__(self, phtrees, epsilon):
        self.phtrees = phtrees
        self.epsilon = epsilon

    def __call__(self, death_index):
        return self.phtrees.nodes[death_index].stable_volume(self.epsilon)

    query_target_name = "stable-volume"


def main(args=None):
    if not args:
        args = argument_parser().parse_args()

    with PDGMReader.open(args.input) as reader:
        assert reader.metadata["filtration_type"] in ["alpha", "alpha-phtrees"]
        input_dim = reader.metadata["dim"]
        pdgm = PDGM(reader, input_dim - 1)
        spatial_searcher = SpatialSearcher(pdgm.death_indices, pdgm.births, pdgm.deaths)
        phtrees = PHTrees.from_pdgm(pdgm)

        if is_point_query(args):
            query_class = PointQuery
            query_xy = (args.x, args.y)
        elif is_rectangle_query(args):
            query_class = RectangleQuery
            query_xy = (args.x_range, args.y_range)

        if args.stable_volume:
            volume_getter = GetStableVolume(phtrees, args.stable_volume)
        else:
            volume_getter = GetOptimalVolume(phtrees)

        query = query_class(query_xy, volume_getter, spatial_searcher, input_dim)

        query.invoke()

        if args.json_output:
            with open(args.json_output, "w") as f:
                json.dump(query.to_dict(), f)

        if args.invoke_paraview:
            with TemporaryDirectory() as tmpdir:
                path = os.path.join(tmpdir, "tmp.vtk")
                drawer = phtrees.coord_resolver.build_paraview_drawer(len(query.result), {})
                for i, r in enumerate(query.result):
                    r.draw_volume(drawer, i)
                drawer.output(path)
                utils.invoke_paraview(path, wait=True)


def is_point_query(args):
    return args.x is not None and args.y is not None


def is_rectangle_query(args):
    return args.x_range is not None and args.y_range is not None


def argument_parser():
    p = argparse.ArgumentParser(description="")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument("-x", type=float, help="birth time of the pair")
    p.add_argument("-y", type=float, help="death time of the pair")
    p.add_argument("-X", "--x-range", type=parse_range, help="birth time of the pair")
    p.add_argument("-Y", "--y-range", type=parse_range, help="death time of the pair")
    p.add_argument("-j", "--json-output", help="output in json format")
    p.add_argument(
        "-P", "--invoke-paraview", default=False, action="store_true", help="invoke paraview for visualization"
    )
    p.add_argument("-S", "--stable-volume", type=float, help="stable volume epsilon")
    p.add_argument("input", help="input pht file")
    return p


if __name__ == "__main__":
    main()



import argparse
import struct
from tempfile import TemporaryDirectory
import numpy as np
import ripser

from homcloud.pdgm_format import (
    PDGMWriter,
    PDChunk,
    SimpleChunk,
    BoundaryMapChunk,
    RipsInformationChunk,
    GraphWeightsChunk,
)
from homcloud.version import __version__
from homcloud.license import add_argument_for_license
from homcloud.utils import load_symbols
from homcloud.argparse_common import parse_bool
import homcloud.dipha as dipha
import homcloud.phat_ext as phat


class DistanceMatrix(object):
    def __init__(self, matrix, upper_dim, upper_value=np.inf, vertex_symbols=None):
        assert matrix.ndim == 2 and matrix.shape[0] == matrix.shape[1]
        self.matrix = matrix
        self.upper_dim = upper_dim
        self.upper_value = upper_value
        self.vertex_symbols = vertex_symbols

    def build_rips_filtration(self, save_graph=False, save_cocycle=False):
        return RipsFiltration(
            self.matrix.astype(np.float32), self.upper_dim, self.upper_value, save_graph, save_cocycle
        )

    def build_simplicial_filtration(self, save_boundary_map=False):
        simplices = sorted(self.all_simplices(), key=lambda x: (x[0], len(x[1])))
        index_to_level = [level for (level, simplex) in simplices]
        index_to_simplex = [simplex for (level, simplex) in simplices]
        simplex_to_index = {simplex: index for (index, (value, simplex)) in enumerate(simplices)}
        if self.vertex_symbols is None:
            vertex_symbols = [str(k) for k in range(self.matrix.shape[0])]
        else:
            vertex_symbols = self.vertex_symbols

        return SimplicialFiltration(
            self, index_to_level, index_to_simplex, simplex_to_index, vertex_symbols, save_boundary_map
        )

    def all_simplices(self):
        npoints = self.matrix.shape[0]

        def value_of_simplex(l, i, value):
            return max(np.max(self.matrix[i, l]), value)

        def iter(l, value):
            if len(l) > self.upper_dim + 1:
                return

            for i in range(l[-1] + 1, npoints):
                newvalue = value_of_simplex(l, i, value)
                if newvalue >= self.upper_value:
                    continue
                l.append(i)
                yield newvalue, tuple(l)
                yield from iter(l, newvalue)
                l.pop()

        for i in range(npoints):
            yield (0.0, (i,))
            yield from iter([i], -np.inf)


class RipsFiltration(object):
    def __init__(self, matrix, upper_dim, upper_value, save_graph, save_cocycle):
        self.matrix = matrix
        self.upper_dim = upper_dim
        self.upper_value = upper_value
        self.save_graph = save_graph
        self.save_cocycle = save_cocycle

    def compute_pdgm(self, f, algorithm=None, save_suppl_info=True):
        algorithm = algorithm or self.favorite_algorithm()

        if algorithm == "ripser":
            self.compute_pdgm_by_ripser(f)
        if algorithm == "dipha":
            self.compute_pdgm_by_dipha(f)

    def compute_pdgm_by_ripser(self, f):
        def check_upper_value_should_be_infinity():
            if self.upper_value != np.inf:
                raise RuntimeError(
                    "upper_value is not supported by ripser now. "
                    "This is because ripser returns a curious "
                    "result when upper_value is given"
                )

        def build_pdchunk(d, pairs):
            births = []
            deaths = []
            ess_births = []
            for k in range(pairs.shape[0]):
                birth = pairs[k, 0]
                death = pairs[k, 1]
                if death == np.inf:
                    ess_births.append(birth)
                else:
                    births.append(birth)
                    deaths.append(death)
            return PDChunk("pd", d, births, deaths, ess_births)

        check_upper_value_should_be_infinity()
        diagrams = ripser.ripser(
            self.matrix,
            self.upper_dim,
            self.upper_value,
            2,  # Z/2Z
            True,  # The first argument is a distane matrix, not a pointcloud
            do_cocycles=self.save_cocycle,
        )

        writer = PDGMWriter(f, "rips", self.upper_dim + 1)

        for d, pairs in enumerate(diagrams["dgms"]):
            writer.append_chunk(build_pdchunk(d, pairs))

        self.append_rips_information_chunk(writer)
        self.append_graph_weights_chunk(writer)

        if self.save_cocycle:
            self.append_cocycle_chunk(writer, diagrams["dgms"], diagrams["cocycles"])

        writer.write()

    def append_graph_weights_chunk(self, writer):
        if not self.save_graph:
            return
        writer.append_chunk(GraphWeightsChunk(self.matrix))

    def append_rips_information_chunk(self, writer):
        writer.append_chunk(RipsInformationChunk(self.num_vertices))

    def append_cocycle_chunk(self, writer, dgms, cocycles):
        for d in range(len(dgms)):
            writer.append_simple_chunk(
                "cocycles", [cocycle.tolist() for (cocycle, pair) in zip(cocycles[d], dgms[d]) if pair[1] != np.inf], d
            )

    @property
    def num_vertices(self):
        return self.matrix.shape[0]

    def compute_pdgm_by_dipha(self, f):
        pd_chunks = [PDChunk("pd", d, [], [], []) for d in range(self.upper_dim + 1)]
        pairs = dipha.pairs_from_filtration(self, 1, False, self.upper_dim, self.upper_value)
        for d, birth, death in pairs:
            if death == np.inf:
                pd_chunks[d].ess_births.append(birth)
            else:
                pd_chunks[d].births.append(birth)
                pd_chunks[d].deaths.append(death)

        writer = PDGMWriter(f, "rips", self.upper_dim + 1)
        writer.extend_chunks(pd_chunks)
        self.append_rips_information_chunk(writer)
        self.append_graph_weights_chunk(writer)
        writer.write()

    def write_dipha_complex(self, output):
        num_vertices = self.matrix.shape[0]

        def write_dipha_header():
            output.write(struct.pack("qqq", 8067171840, 7, num_vertices))

        def write_matrix():
            for k in range(num_vertices):
                for l in range(num_vertices):
                    output.write(struct.pack("d", self.matrix[k, l] * 2))

        write_dipha_header()
        write_matrix()

    def favorite_algorithm(self):
        return "ripser" if self.upper_value == np.inf else "dipha"


class SimplicialFiltration(object):
    def __init__(
        self, dmatrix, index_to_level, index_to_simplex, simplex_to_index, vertex_symbols, save_boundary_map=False
    ):
        self.distance_matrix = dmatrix
        self.index_to_level = index_to_level
        self.index_to_simplex = index_to_simplex
        self.simplex_to_index = simplex_to_index
        self.vertex_symbols = vertex_symbols
        self.boundary_map = self.build_boundary_map()
        self.save_boundary_map = save_boundary_map

    def build_boundary_map(self):
        def boundary(simplex):
            if len(simplex) == 1:
                return

            for k in range(len(simplex)):
                key = simplex[:k] + simplex[k + 1 :]
                yield self.simplex_to_index[key]

        return [[len(simplex) - 1, list(boundary(simplex))] for simplex in self.index_to_simplex]

    @property
    def upper_dim(self):
        return self.distance_matrix.upper_dim

    def build_phat_matrix(self):
        phat_matrix = phat.Matrix(len(self.index_to_simplex), "none")
        for index, (dim, col) in enumerate(self.boundary_map):
            phat_matrix.set_dim_col(index, dim, col)

        return phat_matrix

    def compute_pdgm(self, f, algorithm=None, save_suppl_info=True):
        assert save_suppl_info

        phat_matrix = self.build_phat_matrix()
        phat_matrix.reduce(algorithm)
        writer = PDGMWriter(f, "simplicial", self.upper_dim + 1)
        writer.save_pairs(phat_matrix.birth_death_pairs(), self.index_to_level)
        writer.append_simple_chunk("index_to_level", self.index_to_level)
        writer.append_simple_chunk("index_to_simplex", self.index_to_simplex)
        writer.append_simple_chunk("vertex_symbols", self.vertex_symbols)
        if self.save_boundary_map:
            writer.append_chunk(BoundaryMapChunk("simplicial", self.boundary_map))

        writer.write()


def main(args=None):
    args = args or argument_parser().parse_args()
    matrix = DistanceMatrix(
        np.loadtxt(args.input), args.upper_degree, args.upper_value, load_symbols(args.vertex_symbols)
    )
    if args.save_boundary_map:
        filt = matrix.build_simplicial_filtration(True)
    else:
        filt = matrix.build_rips_filtration()

    with open(args.output, "wb") as f:
        filt.compute_pdgm(f, args.algorithm)


def argument_parser():
    p = argparse.ArgumentParser(description="Compute a PD from Vietris-Rips filtration")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument("-d", "--upper-degree", type=int, required=True, help="Maximum computed degree")
    p.add_argument("-u", "--upper-value", type=float, default=np.inf, help="Maximum distance (default: +inf)")
    p.add_argument("--vertex-symbols", help="vertex symbols file")
    p.add_argument(
        "-M",
        "--save-boundary-map",
        default=False,
        type=parse_bool,
        help="save boundary map into output file" "(only available with phat-* algorithms, on/*off*)",
    )
    p.add_argument("--algorithm", default=None, help="algorithm (dipha, ripser)")
    p.add_argument("--parallels", default=1, type=int, help="number of threads (default: 1)")
    add_argument_for_license(p)
    p.add_argument("input", help="input file")
    p.add_argument("output", help="output file")
    return p


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/plot_PD_GUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(876, 1038)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.mplvl = QtWidgets.QVBoxLayout()
        self.mplvl.setObjectName("mplvl")
        self.horizontalLayout.addLayout(self.mplvl)
        self.frame = QtWidgets.QFrame(self.centralwidget)
        self.frame.setMaximumSize(QtCore.QSize(300, 16777215))
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.frame_optimal_volume = QtWidgets.QFrame(self.frame)
        self.frame_optimal_volume.setMinimumSize(QtCore.QSize(0, 0))
        self.frame_optimal_volume.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_optimal_volume.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_optimal_volume.setObjectName("frame_optimal_volume")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.frame_optimal_volume)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        self.label_13 = QtWidgets.QLabel(self.frame_optimal_volume)
        self.label_13.setObjectName("label_13")
        self.horizontalLayout_13.addWidget(self.label_13)
        self.lineEdit_optimal_volume_stable_volume = QtWidgets.QLineEdit(self.frame_optimal_volume)
        self.lineEdit_optimal_volume_stable_volume.setObjectName("lineEdit_optimal_volume_stable_volume")
        self.horizontalLayout_13.addWidget(self.lineEdit_optimal_volume_stable_volume)
        self.gridLayout_5.addLayout(self.horizontalLayout_13, 7, 0, 1, 2)
        self.radioButton_eigenvolume_tightened = QtWidgets.QRadioButton(self.frame_optimal_volume)
        self.radioButton_eigenvolume_tightened.setObjectName("radioButton_eigenvolume_tightened")
        self.gridLayout_5.addWidget(self.radioButton_eigenvolume_tightened, 5, 1, 1, 1)
        self.radioButton_eigenvolume_tightened_sub = QtWidgets.QRadioButton(self.frame_optimal_volume)
        self.radioButton_eigenvolume_tightened_sub.setObjectName("radioButton_eigenvolume_tightened_sub")
        self.gridLayout_5.addWidget(self.radioButton_eigenvolume_tightened_sub, 6, 1, 1, 1)
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.label_12 = QtWidgets.QLabel(self.frame_optimal_volume)
        self.label_12.setObjectName("label_12")
        self.horizontalLayout_12.addWidget(self.label_12)
        self.lineEdit_optimal_volume_retry = QtWidgets.QLineEdit(self.frame_optimal_volume)
        self.lineEdit_optimal_volume_retry.setMaximumSize(QtCore.QSize(70, 16777215))
        self.lineEdit_optimal_volume_retry.setObjectName("lineEdit_optimal_volume_retry")
        self.horizontalLayout_12.addWidget(self.lineEdit_optimal_volume_retry)
        self.label_11 = QtWidgets.QLabel(self.frame_optimal_volume)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout_12.addWidget(self.label_11)
        self.lineEdit_optimal_volume_cutoff_radius = QtWidgets.QLineEdit(self.frame_optimal_volume)
        self.lineEdit_optimal_volume_cutoff_radius.setObjectName("lineEdit_optimal_volume_cutoff_radius")
        self.horizontalLayout_12.addWidget(self.lineEdit_optimal_volume_cutoff_radius)
        self.gridLayout_5.addLayout(self.horizontalLayout_12, 3, 0, 1, 2)
        self.radioButton_eigenvolume_owned = QtWidgets.QRadioButton(self.frame_optimal_volume)
        self.radioButton_eigenvolume_owned.setObjectName("radioButton_eigenvolume_owned")
        self.gridLayout_5.addWidget(self.radioButton_eigenvolume_owned, 6, 0, 1, 1)
        self.radioButton_eigenvolume_none = QtWidgets.QRadioButton(self.frame_optimal_volume)
        self.radioButton_eigenvolume_none.setChecked(True)
        self.radioButton_eigenvolume_none.setObjectName("radioButton_eigenvolume_none")
        self.gridLayout_5.addWidget(self.radioButton_eigenvolume_none, 5, 0, 1, 1)
        self.checkBox_optimal_volume_query_children = QtWidgets.QCheckBox(self.frame_optimal_volume)
        self.checkBox_optimal_volume_query_children.setObjectName("checkBox_optimal_volume_query_children")
        self.gridLayout_5.addWidget(self.checkBox_optimal_volume_query_children, 1, 0, 1, 2)
        self.checkBox_optimal_volume_on = QtWidgets.QCheckBox(self.frame_optimal_volume)
        self.checkBox_optimal_volume_on.setObjectName("checkBox_optimal_volume_on")
        self.gridLayout_5.addWidget(self.checkBox_optimal_volume_on, 0, 0, 1, 2)
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.checkBox_optimal_volume_show_children = QtWidgets.QCheckBox(self.frame_optimal_volume)
        self.checkBox_optimal_volume_show_children.setChecked(True)
        self.checkBox_optimal_volume_show_children.setObjectName("checkBox_optimal_volume_show_children")
        self.horizontalLayout_11.addWidget(self.checkBox_optimal_volume_show_children)
        self.checkBox_optimal_volume_show_volume = QtWidgets.QCheckBox(self.frame_optimal_volume)
        self.checkBox_optimal_volume_show_volume.setChecked(True)
        self.checkBox_optimal_volume_show_volume.setObjectName("checkBox_optimal_volume_show_volume")
        self.horizontalLayout_11.addWidget(self.checkBox_optimal_volume_show_volume)
        self.gridLayout_5.addLayout(self.horizontalLayout_11, 2, 0, 1, 2)
        self.label_16 = QtWidgets.QLabel(self.frame_optimal_volume)
        self.label_16.setObjectName("label_16")
        self.gridLayout_5.addWidget(self.label_16, 4, 0, 1, 1)
        self.verticalLayout_2.addWidget(self.frame_optimal_volume)
        self.frame_phtree = QtWidgets.QFrame(self.frame)
        self.frame_phtree.setFrameShape(QtWidgets.QFrame.Panel)
        self.frame_phtree.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_phtree.setObjectName("frame_phtree")
        self.gridLayout = QtWidgets.QGridLayout(self.frame_phtree)
        self.gridLayout.setObjectName("gridLayout")
        self.button_show_descendants = QtWidgets.QPushButton(self.frame_phtree)
        self.button_show_descendants.setObjectName("button_show_descendants")
        self.gridLayout.addWidget(self.button_show_descendants, 4, 0, 1, 2)
        self.checkBox_query_volumes_in_rectangle = QtWidgets.QCheckBox(self.frame_phtree)
        self.checkBox_query_volumes_in_rectangle.setObjectName("checkBox_query_volumes_in_rectangle")
        self.gridLayout.addWidget(self.checkBox_query_volumes_in_rectangle, 5, 0, 1, 2)
        self.checkBox_show_descendants = QtWidgets.QCheckBox(self.frame_phtree)
        self.checkBox_show_descendants.setChecked(True)
        self.checkBox_show_descendants.setObjectName("checkBox_show_descendants")
        self.gridLayout.addWidget(self.checkBox_show_descendants, 2, 1, 1, 1)
        self.checkBox_show_ancestors = QtWidgets.QCheckBox(self.frame_phtree)
        self.checkBox_show_ancestors.setChecked(True)
        self.checkBox_show_ancestors.setObjectName("checkBox_show_ancestors")
        self.gridLayout.addWidget(self.checkBox_show_ancestors, 2, 0, 1, 1)
        self.checkBox_query_phtree = QtWidgets.QCheckBox(self.frame_phtree)
        self.checkBox_query_phtree.setObjectName("checkBox_query_phtree")
        self.gridLayout.addWidget(self.checkBox_query_phtree, 1, 0, 1, 2)
        self.verticalLayout_2.addWidget(self.frame_phtree)
        self.frame_3 = QtWidgets.QFrame(self.frame)
        self.frame_3.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setObjectName("frame_3")
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout(self.frame_3)
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.label_10 = QtWidgets.QLabel(self.frame_3)
        self.label_10.setMinimumSize(QtCore.QSize(100, 0))
        self.label_10.setObjectName("label_10")
        self.horizontalLayout_10.addWidget(self.label_10)
        self.radioButton_normalplot = QtWidgets.QRadioButton(self.frame_3)
        self.radioButton_normalplot.setChecked(True)
        self.radioButton_normalplot.setObjectName("radioButton_normalplot")
        self.horizontalLayout_10.addWidget(self.radioButton_normalplot)
        self.radioButton_contourplot = QtWidgets.QRadioButton(self.frame_3)
        self.radioButton_contourplot.setObjectName("radioButton_contourplot")
        self.horizontalLayout_10.addWidget(self.radioButton_contourplot)
        self.verticalLayout_2.addWidget(self.frame_3)
        self.frame_2 = QtWidgets.QFrame(self.frame)
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.frame_2)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.horizontalLayout_14 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_14.setObjectName("horizontalLayout_14")
        self.label_14 = QtWidgets.QLabel(self.frame_2)
        self.label_14.setObjectName("label_14")
        self.horizontalLayout_14.addWidget(self.label_14)
        self.edit_colormap = QtWidgets.QLineEdit(self.frame_2)
        self.edit_colormap.setObjectName("edit_colormap")
        self.horizontalLayout_14.addWidget(self.edit_colormap)
        self.gridLayout_2.addLayout(self.horizontalLayout_14, 6, 0, 1, 3)
        self.label_6 = QtWidgets.QLabel(self.frame_2)
        self.label_6.setObjectName("label_6")
        self.gridLayout_2.addWidget(self.label_6, 0, 0, 1, 1)
        self.radioButton_linear = QtWidgets.QRadioButton(self.frame_2)
        self.radioButton_linear.setChecked(True)
        self.radioButton_linear.setObjectName("radioButton_linear")
        self.gridLayout_2.addWidget(self.radioButton_linear, 1, 0, 1, 1)
        self.radioButton_log = QtWidgets.QRadioButton(self.frame_2)
        self.radioButton_log.setObjectName("radioButton_log")
        self.gridLayout_2.addWidget(self.radioButton_log, 1, 1, 1, 1)
        self.radioButton_loglog = QtWidgets.QRadioButton(self.frame_2)
        self.radioButton_loglog.setObjectName("radioButton_loglog")
        self.gridLayout_2.addWidget(self.radioButton_loglog, 1, 2, 1, 1)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.radioButton_power = QtWidgets.QRadioButton(self.frame_2)
        self.radioButton_power.setObjectName("radioButton_power")
        self.horizontalLayout_7.addWidget(self.radioButton_power)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem1)
        self.label_7 = QtWidgets.QLabel(self.frame_2)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_7.addWidget(self.label_7)
        self.edit_Exp = QtWidgets.QLineEdit(self.frame_2)
        self.edit_Exp.setMaximumSize(QtCore.QSize(128, 16777215))
        self.edit_Exp.setObjectName("edit_Exp")
        self.horizontalLayout_7.addWidget(self.edit_Exp)
        self.gridLayout_2.addLayout(self.horizontalLayout_7, 4, 0, 1, 3)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.label_9 = QtWidgets.QLabel(self.frame_2)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_9.addWidget(self.label_9)
        self.edit_VMax = QtWidgets.QLineEdit(self.frame_2)
        self.edit_VMax.setObjectName("edit_VMax")
        self.horizontalLayout_9.addWidget(self.edit_VMax)
        self.gridLayout_2.addLayout(self.horizontalLayout_9, 5, 0, 1, 3)
        self.verticalLayout_2.addWidget(self.frame_2)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.checkBox_Diffusing = QtWidgets.QCheckBox(self.frame)
        self.checkBox_Diffusing.setObjectName("checkBox_Diffusing")
        self.horizontalLayout_8.addWidget(self.checkBox_Diffusing)
        self.label_8 = QtWidgets.QLabel(self.frame)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_8.addWidget(self.label_8)
        self.edit_DiffusingSize = QtWidgets.QLineEdit(self.frame)
        self.edit_DiffusingSize.setObjectName("edit_DiffusingSize")
        self.horizontalLayout_8.addWidget(self.edit_DiffusingSize)
        self.verticalLayout_2.addLayout(self.horizontalLayout_8)
        self.line = QtWidgets.QFrame(self.frame)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout_2.addWidget(self.line)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_3 = QtWidgets.QLabel(self.frame)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_4.addWidget(self.label_3)
        self.slider_XMin = QtWidgets.QSlider(self.frame)
        self.slider_XMin.setMinimum(0)
        self.slider_XMin.setMaximum(100)
        self.slider_XMin.setSingleStep(1)
        self.slider_XMin.setPageStep(10)
        self.slider_XMin.setProperty("value", 0)
        self.slider_XMin.setOrientation(QtCore.Qt.Horizontal)
        self.slider_XMin.setTickInterval(0)
        self.slider_XMin.setObjectName("slider_XMin")
        self.horizontalLayout_4.addWidget(self.slider_XMin)
        self.edit_XMin = QtWidgets.QLineEdit(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.edit_XMin.sizePolicy().hasHeightForWidth())
        self.edit_XMin.setSizePolicy(sizePolicy)
        self.edit_XMin.setMaximumSize(QtCore.QSize(50, 16777215))
        self.edit_XMin.setObjectName("edit_XMin")
        self.horizontalLayout_4.addWidget(self.edit_XMin)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_2 = QtWidgets.QLabel(self.frame)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_3.addWidget(self.label_2)
        self.slider_XMax = QtWidgets.QSlider(self.frame)
        self.slider_XMax.setMinimum(0)
        self.slider_XMax.setMaximum(100)
        self.slider_XMax.setProperty("value", 100)
        self.slider_XMax.setOrientation(QtCore.Qt.Horizontal)
        self.slider_XMax.setObjectName("slider_XMax")
        self.horizontalLayout_3.addWidget(self.slider_XMax)
        self.edit_XMax = QtWidgets.QLineEdit(self.frame)
        self.edit_XMax.setMaximumSize(QtCore.QSize(50, 16777215))
        self.edit_XMax.setObjectName("edit_XMax")
        self.horizontalLayout_3.addWidget(self.edit_XMax)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.checkBox_yrange = QtWidgets.QCheckBox(self.frame)
        self.checkBox_yrange.setCheckable(True)
        self.checkBox_yrange.setChecked(False)
        self.checkBox_yrange.setObjectName("checkBox_yrange")
        self.verticalLayout_2.addWidget(self.checkBox_yrange)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_5 = QtWidgets.QLabel(self.frame)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_6.addWidget(self.label_5)
        self.slider_YMin = QtWidgets.QSlider(self.frame)
        self.slider_YMin.setMinimum(0)
        self.slider_YMin.setMaximum(100)
        self.slider_YMin.setOrientation(QtCore.Qt.Horizontal)
        self.slider_YMin.setObjectName("slider_YMin")
        self.horizontalLayout_6.addWidget(self.slider_YMin)
        self.edit_YMin = QtWidgets.QLineEdit(self.frame)
        self.edit_YMin.setMaximumSize(QtCore.QSize(50, 16777215))
        self.edit_YMin.setReadOnly(False)
        self.edit_YMin.setObjectName("edit_YMin")
        self.horizontalLayout_6.addWidget(self.edit_YMin)
        self.verticalLayout_2.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_4 = QtWidgets.QLabel(self.frame)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_5.addWidget(self.label_4)
        self.slider_YMax = QtWidgets.QSlider(self.frame)
        self.slider_YMax.setMinimum(0)
        self.slider_YMax.setMaximum(100)
        self.slider_YMax.setProperty("value", 100)
        self.slider_YMax.setOrientation(QtCore.Qt.Horizontal)
        self.slider_YMax.setObjectName("slider_YMax")
        self.horizontalLayout_5.addWidget(self.slider_YMax)
        self.edit_YMax = QtWidgets.QLineEdit(self.frame)
        self.edit_YMax.setMaximumSize(QtCore.QSize(50, 16777215))
        self.edit_YMax.setReadOnly(False)
        self.edit_YMax.setObjectName("edit_YMax")
        self.horizontalLayout_5.addWidget(self.edit_YMax)
        self.verticalLayout_2.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(self.frame)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.edit_XBins = QtWidgets.QLineEdit(self.frame)
        self.edit_XBins.setObjectName("edit_XBins")
        self.horizontalLayout_2.addWidget(self.edit_XBins)
        self.label_15 = QtWidgets.QLabel(self.frame)
        self.label_15.setObjectName("label_15")
        self.horizontalLayout_2.addWidget(self.label_15)
        self.edit_YBins = QtWidgets.QLineEdit(self.frame)
        self.edit_YBins.setObjectName("edit_YBins")
        self.horizontalLayout_2.addWidget(self.edit_YBins)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.buttonReplot = QtWidgets.QPushButton(self.frame)
        self.buttonReplot.setObjectName("buttonReplot")
        self.verticalLayout_2.addWidget(self.buttonReplot)
        self.buttonUndo = QtWidgets.QPushButton(self.frame)
        self.buttonUndo.setObjectName("buttonUndo")
        self.verticalLayout_2.addWidget(self.buttonUndo)
        self.horizontalLayout.addWidget(self.frame)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 876, 20))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.toolBar = QtWidgets.QToolBar(MainWindow)
        self.toolBar.setObjectName("toolBar")
        MainWindow.addToolBar(QtCore.Qt.TopToolBarArea, self.toolBar)
        self.actionOpen = QtWidgets.QAction(MainWindow)
        self.actionOpen.setObjectName("actionOpen")
        self.actionQuit = QtWidgets.QAction(MainWindow)
        self.actionQuit.setObjectName("actionQuit")
        self.actionMeasure = QtWidgets.QAction(MainWindow)
        self.actionMeasure.setCheckable(True)
        self.actionMeasure.setObjectName("actionMeasure")
        self.actionRange = QtWidgets.QAction(MainWindow)
        self.actionRange.setCheckable(True)
        self.actionRange.setObjectName("actionRange")
        self.actionCount = QtWidgets.QAction(MainWindow)
        self.actionCount.setCheckable(True)
        self.actionCount.setObjectName("actionCount")
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menubar.addAction(self.menuFile.menuAction())
        self.toolBar.addAction(self.actionMeasure)
        self.toolBar.addAction(self.actionRange)
        self.toolBar.addAction(self.actionCount)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "plot_PD"))
        self.label_13.setText(_translate("MainWindow", "eigenvolume threshold:"))
        self.lineEdit_optimal_volume_stable_volume.setText(_translate("MainWindow", "0.001"))
        self.radioButton_eigenvolume_tightened.setText(_translate("MainWindow", "tightened"))
        self.radioButton_eigenvolume_tightened_sub.setText(_translate("MainWindow", "tightened sub"))
        self.label_12.setText(_translate("MainWindow", "Retry: "))
        self.lineEdit_optimal_volume_retry.setText(_translate("MainWindow", "4"))
        self.label_11.setText(_translate("MainWindow", "cutoff-radius:"))
        self.lineEdit_optimal_volume_cutoff_radius.setText(_translate("MainWindow", "0.01"))
        self.radioButton_eigenvolume_owned.setText(_translate("MainWindow", "owned"))
        self.radioButton_eigenvolume_none.setText(_translate("MainWindow", "None"))
        self.checkBox_optimal_volume_query_children.setText(_translate("MainWindow", "Query children (slower)"))
        self.checkBox_optimal_volume_on.setText(_translate("MainWindow", "Query volume optimal cycle"))
        self.checkBox_optimal_volume_show_children.setText(_translate("MainWindow", "Show children"))
        self.checkBox_optimal_volume_show_volume.setText(_translate("MainWindow", "Show volume"))
        self.label_16.setText(_translate("MainWindow", "Eigenvolume:"))
        self.button_show_descendants.setText(_translate("MainWindow", "Show descendants volume"))
        self.checkBox_query_volumes_in_rectangle.setText(_translate("MainWindow", "Query volumes in rectangle"))
        self.checkBox_show_descendants.setText(_translate("MainWindow", "Show descendants"))
        self.checkBox_show_ancestors.setText(_translate("MainWindow", "Show ancestors"))
        self.checkBox_query_phtree.setText(_translate("MainWindow", "Query tree"))
        self.label_10.setText(_translate("MainWindow", "Plotting style:"))
        self.radioButton_normalplot.setText(_translate("MainWindow", "&normal"))
        self.radioButton_contourplot.setText(_translate("MainWindow", "&contour"))
        self.label_14.setText(_translate("MainWindow", "colormap: "))
        self.label_6.setText(_translate("MainWindow", "Norm:"))
        self.radioButton_linear.setText(_translate("MainWindow", "&linear"))
        self.radioButton_log.setText(_translate("MainWindow", "lo&g"))
        self.radioButton_loglog.setText(_translate("MainWindow", "lo&glog"))
        self.radioButton_power.setText(_translate("MainWindow", "&power"))
        self.label_7.setText(_translate("MainWindow", "exp:"))
        self.edit_Exp.setText(_translate("MainWindow", "1"))
        self.label_9.setText(_translate("MainWindow", "max:"))
        self.checkBox_Diffusing.setText(_translate("MainWindow", "diffusing pairs"))
        self.label_8.setText(_translate("MainWindow", "size:"))
        self.label_3.setText(_translate("MainWindow", "X min:"))
        self.edit_XMin.setText(_translate("MainWindow", "0"))
        self.label_2.setText(_translate("MainWindow", "X max:"))
        self.edit_XMax.setText(_translate("MainWindow", "0.1"))
        self.checkBox_yrange.setText(_translate("MainWindow", "xrange == yrange"))
        self.label_5.setText(_translate("MainWindow", "Y min:"))
        self.edit_YMin.setText(_translate("MainWindow", "0"))
        self.label_4.setText(_translate("MainWindow", "Y max:"))
        self.edit_YMax.setText(_translate("MainWindow", "0.1"))
        self.label.setText(_translate("MainWindow", "XBins:"))
        self.edit_XBins.setText(_translate("MainWindow", "128"))
        self.label_15.setText(_translate("MainWindow", "YBins: "))
        self.buttonReplot.setText(_translate("MainWindow", "Replot"))
        self.buttonUndo.setText(_translate("MainWindow", "Undo"))
        self.menuFile.setTitle(_translate("MainWindow", "F&ile"))
        self.toolBar.setWindowTitle(_translate("MainWindow", "toolBar"))
        self.actionOpen.setText(_translate("MainWindow", "&Open"))
        self.actionQuit.setText(_translate("MainWindow", "&Quit"))
        self.actionQuit.setShortcut(_translate("MainWindow", "Ctrl+Q"))
        self.actionMeasure.setText(_translate("MainWindow", "Measure"))
        self.actionMeasure.setToolTip(_translate("MainWindow", "Measure distance"))
        self.actionRange.setText(_translate("MainWindow", "Range"))
        self.actionRange.setToolTip(_translate("MainWindow", "Range specification"))
        self.actionCount.setText(_translate("MainWindow", "Count"))



import struct

import numpy as np
import msgpack
from forwardable import forwardable
from cached_property import cached_property

from homcloud.pict.utils import nd_indices
from homcloud.base_filtration import Filtration
import homcloud.homccube as homccube
from homcloud.cubical_ext import CubicalFiltrationExt
from homcloud.pdgm_format import PDGMWriter, SimpleChunk, BitmapInformationChunk, BinaryChunk


class Bitmap(object):
    def __init__(self, array, flip_levels_sign=False, periodicity=None, save_boundary_map=False):
        self.array = array
        self.flip_levels_sign = flip_levels_sign
        self.periodicity = periodicity or [False] * array.ndim
        self.save_boundary_map = save_boundary_map

    def build_bitmap_filtration(self):
        indices = nd_indices(self.array.shape)
        keys = np.argsort(self.array, axis=None, kind="mergesort")
        index_to_level = self.sign * self.array.flatten()[keys]
        index_bitmap = invert_permutation(keys).reshape(self.array.shape)
        return BitmapFiltration(
            index_bitmap, self.periodicity, index_to_level, indices[keys, :], self.flip_levels_sign
        )

    def build_cubical_filtration(self):
        cubefilt_ext = CubicalFiltrationExt(self.array.astype(float), self.periodicity, self.save_boundary_map)
        return CubicalFiltration(cubefilt_ext, self.flip_levels_sign)

    def build_filtration(self, cubical):
        if cubical:
            return self.build_cubical_filtration()
        else:
            return self.build_bitmap_filtration()

    @property
    def sign(self):
        return -1 if self.flip_levels_sign else 1


class BitmapFiltration(Filtration):
    def __init__(self, index_bitmap, periodicity, index_to_level, index_to_pixel, sign_flipped):
        self.index_bitmap = index_bitmap
        self.periodicity = periodicity
        self.index_to_level = index_to_level
        self.index_to_pixel = index_to_pixel
        self.sign_flipped = sign_flipped
        self.dim = index_bitmap.ndim

    def write_dipha_complex(self, f):
        """Write a picture to f with the dipha format."""
        if any(self.periodicity):
            raise NotImplementedError("dipha does not support periodic bitmap")

        f.write(struct.pack("qq", 8067171840, 1))
        f.write(struct.pack("qq", self.index_bitmap.size, self.dim))
        for g in reversed(self.index_bitmap.shape):
            f.write(struct.pack("q", g))

        for pixel in self.index_bitmap.flatten():
            f.write(struct.pack("d", pixel))

    def compute_pdgm(self, f, algorithm=None, save_suppl_info=True):
        writer = PDGMWriter(f, "bitmap", self.dim)
        writer.sign_flipped = self.sign_flipped
        pairs, _ = self.compute_pairs(algorithm)
        writer.save_pairs(pairs, self.index_to_level, save_suppl_info)

        if save_suppl_info:
            writer.append_chunk(SimpleChunk("index_to_level", list(self.index_to_level)))
            writer.append_chunk(SimpleChunk("index_to_pixel", self.index_to_pixel.tolist()))
            writer.append_chunk(BitmapInformationChunk(self.index_bitmap.shape, self.periodicity))
        writer.write()

    def favorite_algorithm(self):
        if self.index_bitmap.ndim in [2, 3]:
            return "homccube-1"
        else:
            return "dipha"


def invert_permutation(a):
    """Create the inverse of the permutation given by a.

    If a is not a permutation, the result is indefinite.
    """
    s = np.zeros(a.size, dtype=np.float64)
    i = np.arange(a.size, dtype=np.int32)
    np.put(s, a, i)
    return s


@forwardable()
class CubicalFiltration(Filtration):
    def_delegators("cubefilt_ext", "value_at")
    def_delegators("cubefilt_ext", "encode_cube")
    def_delegators("cubefilt_ext", "decode_cube")
    def_delegators("cubefilt_ext", "all_cubes")

    def __init__(self, cubefilt_ext, sign_flipped):
        self.cubefilt_ext = cubefilt_ext
        self.sign_flipped = sign_flipped

    @property
    def dim(self):
        return self.cubefilt_ext.array.ndim

    @property
    def shape(self):
        return self.cubefilt_ext.array.shape

    @property
    def periodicity(self):
        return self.cubefilt_ext.periodicity

    @property
    def index_bitmap(self):
        return self.cubefilt_ext.array

    @cached_property
    def index_to_level(self):
        sign = (-1) ** int(self.sign_flipped)
        return sign * self.cubefilt_ext.levels()

    @property
    def index_to_cube(self):
        return self.cubefilt_ext.sorted_cubes

    def build_phat_matrix(self):
        return self.cubefilt_ext.build_phat_matrix()

    def write_dipha_complex(self, output):
        output.write(self.cubefilt_ext.dipha_byte_sequence())

    @staticmethod
    def favorite_algorithm():
        return "phat-twist"

    def compute_pdgm(self, f, algorithm=None, save_suppl_info=True):
        writer = PDGMWriter(f, "cubical", self.dim)
        pairs, boundary_map_byte_sequence = self.compute_pairs(algorithm)
        writer.save_pairs(pairs, self.index_to_level, save_suppl_info)

        if save_suppl_info:
            writer.append_simple_chunk("index_to_level", self.index_to_level.tolist())
            writer.append_simple_chunk("index_to_cube", self.index_to_cube)
            writer.append_chunk(BitmapInformationChunk(self.shape, self.periodicity))
        if boundary_map_byte_sequence:
            writer.append_chunk(BinaryChunk("boundary_map", boundary_map_byte_sequence))
        writer.write()



import math

import numpy as np
import matplotlib.colors as colors


class LogPlusOneNorm(colors.Normalize):
    """log(x+1) normalizer for matplotlib.colors."""

    def __init__(self, vmax, vmin=0, clip=False):
        colors.Normalize.__init__(self, vmax=vmax, vmin=vmin, clip=clip)

    def __call__(self, value, clip=None):
        if clip is None:
            clip = self.clip

        result, is_scalar = self.process_value(value)

        self.autoscale_None(result)
        vmax = float(self.vmax)
        if vmax <= 0.0:
            vmax = 1.0
        if clip:
            mask = np.ma.getmask(result)
            result = np.ma.array(np.clip(result.filled(vmax), 0.0, vmax), mask=mask)
        result = np.ma.array(np.log(result.data + 1.0) / math.log(vmax + 1.0), mask=result.mask, copy=False)
        if is_scalar:
            return result[0]
        else:
            return result

    def inverse(self, value):
        if not self.scaled():
            raise ValueError("Not invertible until scaled")

        vmax = float(self.vmax)
        if np.iterable(value):
            return np.ma.exp(value * math.log(vmax + 1)) - 1
        else:
            return math.exp(value * math.log(vmax + 1)) - 1


class LogLogNorm(colors.Normalize):
    """log(log(x+1)+1) normalizer for matplotlib.colors."""

    @staticmethod
    def loglog(x):
        return math.log(math.log(x + 1) + 1)

    def __call__(self, value, clip=None):
        if clip is None:
            clip = self.clip

        result, is_scalar = self.process_value(value)
        vmin, vmax = self.vmin, self.vmax

        if clip:
            result = np.ma.array(np.clip(result.filled(vmax), vmin, vmax), mask=np.ma.getmask(result))
        data = result.data
        data += 1.0
        np.log(data, data)
        data += 1.0
        np.log(data, data)
        data -= self.loglog(vmin)
        data /= self.loglog(vmax) - self.loglog(vmin)
        result = np.ma.array(data, mask=result.mask, copy=False)

        if is_scalar:
            return result[0]
        else:
            return result

    def inverse(self, value):
        if not self.scaled():
            raise ValueError("Not invertible until scaled")

        vmin = float(self.vmin)
        vmax = float(self.vmax)
        loglog_value = self.loglog(vmin) + value * (self.loglog(vmax) - self.loglog(vmin))
        if np.iterable(value):
            return np.ma.exp(np.ma.exp(loglog_value) - 1) - 1
        else:
            return math.exp(math.exp(loglog_value) - 1) - 1


class MidpointNormalize(colors.Normalize):
    """
    This class comes from http://matplotlib.org/users/colormapnorms.html
    """

    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))



from collections import defaultdict
import argparse
import sys

import numpy as np
import msgpack

from homcloud.pdgm_format import PDGMWriter, SimpleChunk, AlphaInformationChunk
from homcloud.version import __version__
from homcloud.license import add_argument_for_license
from homcloud.utils import load_symbols


def coboundary_map(input_dim, boundary_map):
    cobmap = defaultdict(list)
    for index, (d, boundary) in enumerate(boundary_map):
        if d == input_dim:
            for b in boundary:
                cobmap[b].append(index)
    return cobmap


class PHTrees(object):
    def __init__(self, input_dim, boundary_map):
        self.input_dim = input_dim
        self.coboundary_map = coboundary_map(input_dim, boundary_map)
        self.nodes = dict()
        self.infinity = Node(np.inf)
        self.build()

    def node(self, index):
        if index in self.nodes:
            return self.nodes[index]
        else:
            self.nodes[index] = Node(index)
            return self.nodes[index]

    def root(self, node):
        if node.shortcut is None:
            return node
        root = self.root(node.shortcut)
        node.shortcut = root
        return root

    def build(self):
        def merge(root1, root2):
            if root1.death_index == root2.death_index:
                pass
            elif root1.death_index < root2.death_index:
                root1.birth_index = facet
                root1.parent = root2.death_index
                root1.shortcut = root2
            else:
                root2.birth_index = facet
                root2.parent = root1.death_index
                root2.shortcut = root1

        for facet, coboundary in sorted(self.coboundary_map.items(), reverse=True):
            if len(coboundary) == 2:
                merge(self.root(self.node(coboundary[0])), self.root(self.node(coboundary[1])))
            else:
                merge(self.root(self.node(coboundary[0])), self.infinity)

    def to_list(self):
        return [node.to_list() for node in self.nodes.values()]

    def all_pairs(self):
        return [node.to_pair(self.input_dim) for node in self.nodes.values()]

    def save_pdgm(self, f, filt):
        writer = PDGMWriter(f, "alpha-phtrees", self.input_dim)
        writer.save_pairs(self.all_pairs(), filt.index_to_level)
        writer.append_chunk(filt.alpha_information_chunk())
        writer.append_chunk(SimpleChunk("index_to_level", filt.index_to_level))
        writer.append_chunk(SimpleChunk("vertex_symbols", filt.symbols))
        writer.append_chunk(SimpleChunk("vertex_coordintes", filt.points.tolist()))
        writer.append_chunk(SimpleChunk("index_to_simplex", filt.index_to_simplex))
        writer.append_chunk(SimpleChunk("phtrees", self.to_list()))
        writer.write()


class Node(object):
    def __init__(self, death_index):
        self.death_index = death_index
        self.birth_index = None
        self.shortcut = None
        self.parent = None

    def to_list(self):
        return [self.birth_index, self.death_index, self.parent]

    def to_pair(self, dim):
        return (dim - 1, self.birth_index, self.death_index)


def argument_parser():
    p = argparse.ArgumentParser(description="Comput a PH trees from a point cloud")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument("-d", "--dimension", type=int, required=True, help="dimension of the input data")
    p.add_argument("--vertex-symbols", help="vertex symbols file")
    p.add_argument("-w", "--weighted", action="store_true", default=False, help="use an weighted alpha filtration")
    p.add_argument(
        "--no-square",
        action="store_true",
        default=False,
        help="no squared output, if a birth radius is negative, the output is -sqrt(abs(r))",
    )
    p.add_argument("--square", action="store_const", const=False, dest="no_square")
    p.add_argument("--input-type", help="input type (pdgm/pointcloud)")
    add_argument_for_license(p)
    p.add_argument("input")
    p.add_argument("output")
    return p


def main(args=None):
    from homcloud.alpha_filtration import AlphaShape

    args = args or argument_parser().parse_args()
    pointcloud = np.loadtxt(args.input)

    alpha_shape = AlphaShape.build(
        pointcloud,
        args.dimension,
        args.weighted,
    )
    filt = alpha_shape.create_filtration(args.no_square, load_symbols(args.vertex_symbols), True)
    matrix = filt.build_phat_matrix()
    boundary_map = msgpack.loads(matrix.boundary_map_byte_sequence(), raw=False)

    trees = PHTrees(args.dimension, boundary_map["map"])
    with open(args.output, "wb") as f:
        trees.save_pdgm(f, filt)


if __name__ == "__main__":
    sys.exit(main())



import argparse
import sys

import numpy as np

import homcloud.pict.utils as utils
from homcloud.version import __version__
from homcloud.license import add_argument_for_license
from homcloud.argparse_common import (
    parse_cubical_periodic_flags,
    parse_bool,
    check_abolished_output,
    check_cubical_option,
)
from homcloud.bitmap import Bitmap
from .distance_transform import distance_transform


def argument_parser():
    """Return ArgumentParser object to parse command line arguments"""
    p = argparse.ArgumentParser()
    p.description = "Create N-dimensional erosion-dilation filtration "
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument(
        "-T", "--type", default="text_nd", help="input data format (text2d,text_nd(default),picture2d,pictures3d,npy)"
    )
    add_argument_for_binarize_threshold(p)
    p.add_argument("-o", "--output", required=True, help="output complex file name")
    p.add_argument("-s", "--ensmall", action="store_true", default=False, help="ensmall binarized picture")
    p.add_argument(
        "--metric",
        default="manhattan",
        help="metric used to enlarge binarized image " "(manhattan(default), euclidean, etc.)",
    )
    p.add_argument(
        "-C", "--cubical", default=False, action="store_true", help="use explicit cubical filtration (slow)"
    )
    p.add_argument(
        "-p",
        "--periodic",
        default=None,
        type=parse_cubical_periodic_flags,
        help="periodic (example: 0,1,1 for y and z are periodic)",
    )
    p.add_argument("--mask", help="mask bitmap")
    p.add_argument("--algorithm", default=None, help="algorithm (dipha, phat-twist, phat-chunk-parallel)")
    p.add_argument(
        "-M",
        "--save-boundary-map",
        default=False,
        type=parse_bool,
        help="save boundary map into idiagram file" "(only available with phat-* algorithms, on/*off*)",
    )
    add_argument_for_license(p)
    p.add_argument("input", nargs="*", help="input files")
    return p


def add_argument_for_binarize_threshold(p):
    p.add_argument(
        "-m",
        "--mode",
        default="black-base",
        help="the way to construct dipha complex (black-base|white-base, " + "default is black-base)",
    )
    p.add_argument("-t", "--threshold", type=float, default=128, help="threshold for binarization (default: 128)")
    p.add_argument("--gt", type=float, default=None, help="lower threshold")
    p.add_argument("--lt", type=float, default=None, help="upper threshold")


def main(args=None):
    args = args or argument_parser().parse_args()
    check_abolished_output(args)
    check_cubical_option(args)

    pict = utils.load_nd_picture(args.input, args.type)
    binary_pict = binarize_picture(pict, args.threshold, args.mode, (args.gt, args.lt))

    mask = load_mask(args.mask, args.type)
    distance_map = distance_transform(binary_pict, args.metric, args.ensmall, obstacle=mask)
    bitmap = Bitmap(distance_map, False, args.periodic, args.save_boundary_map)
    if args.cubical:
        filt = bitmap.build_cubical_filtration()
    else:
        filt = bitmap.build_bitmap_filtration()

    filt.compute_pdgm_and_save(args.output, args.algorithm)

    return 0


def load_mask(path, type):
    if path:
        return utils.load_nd_picture([path], type) > 0
    else:
        return None


def binarize_picture(pict, threshold, mode, bounds):
    """Binarize pict using threshold.

    Args:
    pict -- a gray-scale picture (ndarray)
    threshold -- threshold for binarization
    mode -- "black-base" or "white-base"
    bounds -- a pair of lower and upper bounds
        if both of them are None, the pair of threshold and mode is used
    """
    lower_bound, upper_bound = bounds

    if lower_bound is None and upper_bound is None:
        if mode == "black-base":
            upper_bound = threshold
        elif mode == "white-base":
            lower_bound = threshold
        else:
            raise RuntimeError("Unknown mode {0}".format(mode))

    lower_bound = -float("inf") if lower_bound is None else lower_bound
    upper_bound = float("inf") if upper_bound is None else upper_bound
    return (lower_bound <= pict) & (pict <= upper_bound)


def mahalanobis_vi(weight, angle):
    def rotation_matrix(t):
        from math import cos, sin

        return np.array([[cos(t), sin(t)], [-sin(t), cos(t)]])

    return rotation_matrix(angle) @ np.diag([weight, 1 / weight]) @ rotation_matrix(-angle)


if __name__ == "__main__":
    sys.exit(main())



import math
import struct
from collections import defaultdict
from operator import attrgetter

import numpy as np
from cached_property import cached_property
import msgpack

from homcloud.alpha_shape2 import AlphaShape2 as AlphaShape2Base
from homcloud.alpha_shape3 import AlphaShape3 as AlphaShape3Base
from homcloud.periodic_alpha_shape3 import PeriodicAlphaShape3 as PeriodicAlphaShape3Base
from homcloud.base_filtration import Filtration
from homcloud.pdgm_format import PDGMWriter, BinaryChunk, AlphaInformationChunk
import homcloud.phat_ext as phat
import homcloud.build_phtrees as build_phtrees


class AlphaShape(object):
    @staticmethod
    def build(points, dim, weighted=False, use_relative_homology=False, periodicity=None):
        if dim == 2:
            return AlphaShape2(points, weighted, use_relative_homology)
        if dim == 3:
            if periodicity is None:
                return AlphaShape3(points, weighted, use_relative_homology)
            else:
                return PeriodicAlphaShape3(points, weighted, use_relative_homology, periodicity)

        raise ValueError("dimension of a point cloud must be 2 or 3")

    def create_filtration(self, squared=True, symbols=None, save_boundary_map=False, save_phtrees=False):
        simplices = [
            Simplex(alpha_simplex, i, squared)
            for (i, alpha_simplex) in enumerate(sorted(self.simplices, key=attrgetter("birth_radius")))
        ]

        dict_simplices = {simplex.key: simplex for simplex in simplices}
        return AlphaFiltration(
            self.coordinates,
            self.weights,
            squared,
            self.periodicity,
            simplices,
            dict_simplices,
            self.dim,
            symbols,
            save_boundary_map,
            save_phtrees,
        )

    def subsets(self):
        def group_of_simplex(simplex):
            """
            If all vertices of the simplex belong to the same group,
            returns an integer which is the name of the group.
            Otherwise, returns None.
            """
            group_names = [v.group_name for v in simplex.vertices()]
            if group_names[0] == -1:
                return None
            if all(gn == group_names[0] for gn in group_names):
                return group_names[0]
            else:
                return None

        def group_by(collection, key):
            groups = defaultdict(list)
            for item in collection:
                k = key(item)
                if k is not None:
                    groups[k].append(item)
            return groups

        return {
            group_name: AlphaSubset(group_name, self.coordinates, self.weights, self.periodicity, simplices, self.dim)
            for (group_name, simplices) in group_by(self.simplices, group_of_simplex).items()
        }

    def all_subsets_acyclic(self):
        return all(subset.isacyclic() for subset in self.subsets().values())

    def check_subsets_acyclicity(self):
        for subset in self.subsets().values():
            if not subset.isacyclic():
                message = "Subset {} is not acyclic".format(subset.group_name)
                raise RuntimeError(message)

    def become_partial_shape(self):
        for subset in self.subsets().values():
            for simplex in subset.simplices:
                simplex.birth_radius = -math.inf

    @staticmethod
    def simplex_key(simplex):
        return tuple(sorted(v.vertex_index for v in simplex.vertices()))


class AlphaShape3(AlphaShape3Base, AlphaShape):
    def __init__(self, points, weighted, rel_homology):
        super().__init__(points, weighted, rel_homology)
        self.weights = points[:, 3] if weighted else None
        self.coordinates = points[:, 0:3]
        self.simplices = self.vertices() + self.edges() + self.facets() + self.cells()

    @property
    def dim(self):
        return 3

    @property
    def periodicity(self):
        return None


class AlphaShape2(AlphaShape2Base, AlphaShape):
    def __init__(self, points, weighted, rel_homology):
        super().__init__(points, weighted, rel_homology)
        self.weights = points[:, 3] if weighted else None
        self.coordinates = points[:, 0:2]
        self.simplices = self.vertices() + self.edges() + self.faces()

    @property
    def dim(self):
        return 2

    @property
    def periodicity(self):
        return None


class PeriodicAlphaShape3(PeriodicAlphaShape3Base, AlphaShape):
    def __init__(self, points, weighted, rel_homology, periodicity):
        ((xmin, xmax), (ymin, ymax), (zmin, zmax)) = periodicity
        super().__init__(points, weighted, rel_homology, xmin, xmax, ymin, ymax, zmin, zmax)
        self.coordinates = points[:, 0:3]
        self.weights = points[:, 3] if weighted else None
        self.simplices = self.vertices() + self.edges() + self.facets() + self.cells()
        self.periodicity = periodicity

    @property
    def dim(self):
        return 3


class AlphaSubset(AlphaShape):
    def __init__(self, group_name, points, weights, periodicity, simplices, dim):
        self.group_name = group_name
        self.periodicity = periodicity
        self.simplices = simplices
        self.coordinates = points
        self.weights = weights
        self.dim = dim

    def isacyclic(self):
        return self.create_filtration().isacyclic()


class Simplex(object):
    """A class representing simplex"""

    def __init__(self, alpha_simplex, index, squared=True):
        self.index = index  # NOTE: This index is different from vertex's index
        self.key = AlphaShape.simplex_key(alpha_simplex)
        self.birth_radius = self.normalize_radius(alpha_simplex.birth_radius, squared)
        self.isvertex = alpha_simplex.isvertex()

    def __repr__(self):
        return "alpha_filtration.Simplex(index={}, key={}, birth_radius={})".format(
            self.index, self.key, self.birth_radius
        )

    def boundary_keys(self):
        """Return list of frozensets of vertices of indices of boundary simplices"""
        if self.isvertex:
            return []
        return [self.key[0:n] + self.key[n + 1 :] for n in range(len(self.key))]

    def signed_boundary_keys(self):
        def plusminus_alternative(length):
            return [(-1 if k % 2 else 1) for k in range(length)]

        unsigned = self.boundary_keys()
        return list(zip(plusminus_alternative(len(unsigned)), unsigned))

    @staticmethod
    def normalize_radius(r, squared):
        return r if squared else math.copysign(math.sqrt(abs(r)), r)

    @property
    def dim(self):
        """Return the dimension of the simplex"""
        return len(self.key) - 1


class AlphaFiltration(Filtration):
    def __init__(
        self,
        points,
        point_weights,
        squared,
        periodicity,
        simplices,
        dict_simplices,
        dim,
        symbols,
        save_boundary_map,
        save_phtrees,
    ):
        """
        Args:
        points -- list of N-d point coordinates
        point_weights -- weights of points, None if weight is not given
        squared -- whether all radii are squared
        periodicity -- None or list of (min, max)
        simplices -- list of simplices, must be sorted by their birth_radius
        dict_simplices -- dictiorary: simplex.key -> simplex
        dim: -- dimension of the alpha filtration
        symbols -- list of symbols, or None
        save_boundary_map -- bool
        save_phTrees -- bool
        """
        self.points = points
        self.point_weights = point_weights
        self.squared = squared
        self.periodicity = periodicity
        self.simplices = simplices
        self.dict_simplices = dict_simplices
        self.dim = dim
        self.symbols = symbols
        self.save_boundary_map = save_boundary_map
        self.save_phtrees = save_phtrees

    def alpha_information_chunk(self):
        return AlphaInformationChunk(self.points.shape[0], self.periodicity, self.points_weighted, self.squared)

    def compute_pdgm(self, f, algorithm=None, output_suppl_info=True):
        writer = PDGMWriter(f, "alpha", self.dim)
        pairs, boundary_map_byte_sequence = self.compute_pairs(algorithm)
        writer.save_pairs(pairs, self.index_to_level, output_suppl_info)
        writer.append_chunk(self.alpha_information_chunk())

        if output_suppl_info:
            writer.append_simple_chunk("index_to_level", self.index_to_level)
            writer.append_simple_chunk("vertex_symbols", self.symbols)
            writer.append_simple_chunk("vertex_coordintes", self.points.tolist())
            writer.append_simple_chunk("index_to_simplex", self.index_to_simplex)

        if self.save_boundary_map:
            writer.append_chunk(BinaryChunk("boundary_map", boundary_map_byte_sequence))
        if self.save_phtrees:
            boundary_map = msgpack.unpackb(boundary_map_byte_sequence, raw=False).get("map")
            writer.append_simple_chunk("phtrees", self.build_phtrees(boundary_map))
        writer.write()

    @cached_property
    def index_to_level(self):
        return [simplex.birth_radius for simplex in self.simplices]

    @cached_property
    def index_to_simplex(self):
        return [list(simplex.key) for simplex in self.simplices]

    @property
    def points_weighted(self):
        return self.point_weights is not None

    def build_phat_matrix(self):
        matrix = phat.Matrix(len(self.simplices), self.boundary_map_style())
        for simplex in self.simplices:
            boundary = [self.dict_simplices[key].index for key in simplex.boundary_keys()]
            matrix.set_dim_col(simplex.index, simplex.dim, boundary)

        return matrix

    def boundary_map_style(self):
        return "simplicial" if self.save_boundary_map else "none"

    def build_phtrees(self, boundary_map):
        return build_phtrees.PHTrees(self.dim, boundary_map).to_list()

    def isacyclic(self):
        matrix = self.build_phat_matrix()
        matrix.reduce_twist()
        return self.count_essential_pairs(matrix.birth_death_pairs()) == 1

    @staticmethod
    def count_essential_pairs(pairs):
        return sum([1 for pair in pairs if pair[2] is None])

    @staticmethod
    def favorite_algorithm():
        return "phat-twist"

    def write_dipha_complex(self, output, use_index=True):
        """ "Write dipha bondary matrix data to file"""

        def write_dipha_header(num_simplices):
            """Write magic number, type, boundary/coboundary,
            num_simplices, and dimension
            """
            output.write(struct.pack("qqqqq", 8067171840, 0, 0, num_simplices, self.dim))

        def write_dimensions():
            """Write dimensions of each simplex"""
            for simplex in self.simplices:
                output.write(struct.pack("q", simplex.dim))

        def write_birth_radii():
            """Write birth radii for each simplex"""
            for simplex in self.simplices:
                if use_index:
                    output.write(struct.pack("d", simplex.index))
                else:
                    output.write(struct.pack("d", simplex.birth_radius))

        def write_boundary_map_sizes():
            n = 0
            for simplex in self.simplices:
                output.write(struct.pack("q", n))
                n += len(simplex.boundary_keys())
            output.write(struct.pack("q", n))  # Write the number of all boundary elements

        def write_boundary_map():
            for simplex in self.simplices:
                for key in simplex.boundary_keys():
                    output.write(struct.pack("q", self.dict_simplices[key].index))

        write_dipha_header(len(self.simplices))
        write_dimensions()
        write_birth_radii()
        write_boundary_map_sizes()
        write_boundary_map()



import operator

from forwardable import forwardable

import homcloud.optvol as optvol
import homcloud.paraview_interface as pv_interface
import homcloud.plotly_3d as p3d


@forwardable()
class Volume(object):
    """
    This class represents a volume.
    This is the superclass of OptimalVolume, StableVolume and
    StableSubvolume.

    Notes:
        * point: list of float
        * cell: simplex or cube, simplex is used if the filtration is
          simplicial (alpha filtration) and cube is used if the filtration
          is cubical.
        * simplex: list of point
        * cube: tuple[point, list of {0, 1}],
        * ssimplex: list of string
    """

    def __init__(self, pair, result):
        self.pair = pair
        self.diagram = pair.diagram
        self.result = result

    def birth_time(self):
        """
        Returns:
            float: The birth time.
        """
        return self.pair.birth_time()

    def death_time(self):
        """
        Returns:
            float: The death time.
        """
        return self.pair.death_time()

    def lifetime(self):
        """
        Returns:
            float: The lifetime of the pair.
        """
        return self.death_time() - self.birth_time()

    def death_position(self):
        """
        Returns:
            simplex: The death simplex.
        """
        return self.pair.death_position

    @property
    def geom_resolver(self):
        return self.diagram.pd.geometry_resolver(False)

    @property
    def symbol_resolver(self):
        return self.diagram.pd.geometry_resolver(True)

    def points(self):
        """
        Returns:
            list of point: All vertices in the optimal volume.
        """
        return self.geom_resolver.resolve_vertices(self.result.cell_indices)

    def points_symbols(self):
        """
        Returns:
            list of string: All vertices in the optimal volume
            in the form of the symbolic representation.
        """
        return self.symbol_resolver.resolve_vertices(self.result.cell_indices)

    def boundary_points(self):
        """
        Returns:
            list of point: All vertices in the volume optimal cycle.
        """
        return self.geom_resolver.resolve_vertices(self.boundary_cells())

    def boundary_points_symbols(self):
        """
        Returns:
            list of string: All vertices in the volume optimal cycle
            in the form of the symbolic representation.
        """
        return self.symbol_resolver.resolve_vertices(self.boundary_cells())

    def boundary_cells(self):
        return self.geom_resolver.boundary_cells(self.result.cell_indices)

    def boundary(self):
        """
        Returns:
            list of cells: All cells in the volume optimal cycle.
        """
        return self.geom_resolver.resolve_cells(self.boundary_cells())

    def boundary_symbols(self):
        """
        Returns:
            list of ssimplex: All simplices in the volume optimal cycle
            in the form of the symbolic representation.
        """
        return self.symbol_resolver.resolve_cells(self.boundary_cells())

    def cells(self):
        """
        Returns:
            list of cell: All cells in volume optimal cycles.
        """
        return self.geom_resolver.resolve_cells(self.result.cell_indices)

    def simplices(self):
        """
        Returns:
            list of simplex: All simplices in volume optimal cycles.
        """
        return self.cells()

    def simplices_symbols(self):
        """
        Returns:
            list of ssimplex: All simplices in volume optimal cycles
            in the form of the symbolic representation.
        """
        return self.symbol_resolver.resolve_cells(self.result.cell_indices)

    volume_simplices_symbols = simplices_symbols

    def cubes(self):
        """
        Returns:
            list of cube: All cubes in volume optimal cycles.
        """
        return self.cells()

    def children(self):
        """
        Returns:
           list of :class:`Pair`: All children pairs.
        """
        from .pd import Pair

        death_to_number = self.diagram.pd.death_index_to_pair_number

        def valid(d):
            return d != self.pair.death_index and d in death_to_number

        return [Pair(self.diagram, death_to_number[d]) for d in self.result.cell_indices if valid(d)]

    def to_dict(self):
        """
        Returns:
            dict: The information about the optimal volume.
        """
        raise NotImplementedError("hc.interface.Pair.to_dict")

    #: The alias of :meth:`death_position`.
    death_pos = death_position

    def to_paraview_node(self, gui_name=None):
        """
        Construct a :class:`homcloud.paraview_interface.PipelineNode` object
        to visulize an optimal volume.

        You can show the optimal volume by
        :meth:`homcloud.paraview_interface.show`. You can also
        adjust the visual by the methods of
        :class:`homcloud.paraview_interface.PipelineNode`.

        Args:
            gui_name (string or None): The name shown in Pipeline Browser
                in paraview's GUI.

        Returns:
            homcloud.paraview_interface.PipelineNode: A PipelineNode object.
        """
        return OptimalVolume.to_paraview_node_for_volumes([self], gui_name)

    to_pvnode = to_paraview_node

    @staticmethod
    def to_paraview_node_for_volumes(volumes, gui_name=None):
        """
        Construct a :class:`homcloud.paraview_interface.PipelineNode` object
        to visulize multiple optimal volumes.

        All optimal volumes should come from the same :class:`PD` object.

        Args:
            volumes (list of :class:`OptimalVolume`): The optimal volumes to be
                visualized.
            gui_name (string or None): The name shown in Pipeline Browser
                in paraview's GUI.

        Returns:
            homcloud.paraview_interface.PipelineNode: A PipelineNode object.
        """
        diagram = volumes[0].diagram
        if diagram.filtration_type == "alpha":
            geom_resolver = diagram.pd.alpha_coord_resolver
            drawer = optvol.drawer_for_alpha(geom_resolver.vertices)
        elif diagram.filtration_type == "cubical":
            geom_resolver = diagram.pd.cubical_geometry_resolver
            drawer = optvol.drawer_for_cubical(geom_resolver.shape)
        optvol.draw_volumes(drawer, map(operator.attrgetter("result"), volumes), geom_resolver)
        f = pv_interface.TempFile(".vtk")
        drawer.write(f)
        f.close()
        return pv_interface.VTK(f.name, gui_name, f).set_representation("Wireframe")

    def to_plotly3d_trace(self, color="green", width=1, name=""):
        """
        Constructs a plotly's trace object to visualize the optimal volume

        Args:
            color (string or None): The name of the color
            width (int): The width of the line
            name (string): The name of the object

        Returns:
            plotly.graph_objects.Scatter3d: Plotly's trace object
        """
        if self.diagram.filtration_type == "alpha":
            return p3d.Simplices(self.boundary(), color, width, name)
        elif self.diagram.filtration_type == "cubical":
            return p3d.Cubes(self.boundary(), color, width, name)
        else:
            raise RuntimeError(f"{self.filtration_type} cannot be renderred")

    to_plotly3d = to_plotly3d_trace

    def to_plotly3d_mesh(self, color="green", name=""):
        """
        Constructs a plotly's trace object to visualize the face of an optimal volume

        Args:
            color (string or None): The name of the color
            name (string): The name of the object

        Returns:
            plotly.graph_objects.Mesh3d: Plotly's trace object
        """
        if self.diagram.filtration_type == "alpha":
            if self.diagram.degree == 2:
                return p3d.SimplicesMesh(self.boundary(), color, name)
            elif self.diagram.degree == 1:
                return p3d.SimplicesMesh(self.simplices(), color, name)
            else:
                raise RuntimeError(f"dim {self.diagram.degree} volume is available for plotly")
        elif self.diagram.filtration_type == "cubical":
            if self.diagram.degree == 2:
                return p3d.CubesMesh(self.boundary(), color, name)
        else:
            raise RuntimeError(f"{self.filtration_type} cannot be renderred")

    def to_pyvista_boundary_mesh(self):
        """
        Constructs a PyVista's mesh object to visualize the boundary of an optimal/stable volume.

        Returns:
            pyvista.PolyData: PyVista's mesh object
        """
        import homcloud.pyvistahelper as pvhelper

        if self.diagram.filtration_type == "alpha":
            if self.diagram.degree == 2:
                return pvhelper.Triangles(self.boundary())
            elif self.diagram.degree == 1:
                return pvhelper.Lines(self.boundary())
            else:
                raise RuntimeError(f"dim {self.diagram.degree} volume is available for pyvista")

        raise RuntimeError(f"{self.filtration_type} cannot be renderred")

    def to_pyvista_volume_mesh(self):
        """
        Constructs a PyVista's mesh object to visualize the internal face of a 1D optimal/stable volume.

        Returns:
            pyvista.PolyData: PyVista's mesh object
        """
        import homcloud.pyvistahelper as pvhelper

        if self.diagram.filtration_type == "alpha":
            if self.diagram.degree == 1:
                return pvhelper.Triangles(self.cells())
        raise RuntimeError(f"Volume mesh can only be applied to 1d alpha volume")


@forwardable()
class OptimalVolume(Volume):
    """
    This class represents an optimal volume.
    """

    def birth_position(self):
        """
        Returns:
            simplex: The birth simplex.
        """
        return self.pair.birth_position

    #: The alias of :meth:`birth_position`.
    birth_pos = birth_position

    def stable_subvolume(self, threshold, solver=None, solver_options=[]):
        """
        Returns the stable subvolume of the optimal volume.

        Args:
            threshold (float): The noise bandwidth.
        Returns:
            StableSubvolume: The stable subvolume.
        """
        lp_solver = optvol.find_lp_solver(solver, solver_options)
        ssvfinder = optvol.TightenedSubVolumeFinder(
            self.diagram.optvol_optimizer_builder(None, None, lp_solver), self.diagram.pd.index_to_level, threshold
        )
        result = ssvfinder.find(self.pair.birth_index, self.pair.death_index, self.result.cell_indices)
        return StableSubvolume(self.pair, result, threshold)

    tightened_subvolume = stable_subvolume


@forwardable()
class EigenVolume(Volume):
    """
    This class represents an "eigenvolume". It is the superclass of
    StableVolume and StableSubvolume.

    Attributes:
        threshold (float): The threshold used for the computation of the
            eigenvolume.
    """

    def __init__(self, pair, result, threshold):
        super().__init__(pair, result)
        self.threshold = threshold


class StableVolume(EigenVolume):
    """
    This class represents a stable volume.

    The instance is given by :meth:`Pair.stable_volume`.
    """

    pass


class StableSubvolume(EigenVolume):
    """
    This class represents a stable subvolume.

    The instance is given by :meth:`OptimalVolume.stable_subvolume`.
    """

    pass



import argparse
import json
import sys
import math

import numpy as np

import homcloud.utils as utils
from homcloud.histogram import Ruler, PDHistogram
from homcloud.version import __version__
import homcloud.pdgm as pdgm
from homcloud.argparse_common import add_arguments_for_load_diagrams, add_arguments_for_histogram_rulers


def argument_parser():
    p = argparse.ArgumentParser(description="Create a finite dim vector from PD")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_arguments_for_load_diagrams(p)
    add_arguments_for_histogram_rulers(p)
    p.add_argument("-D", "--gaussian-sd", type=float, required=True, help="standard deviation of gaussian diffusion")
    p.add_argument("-C", type=float, help="weight constant C")
    p.add_argument("-p", type=float, default=1.0, help="weight constant p")
    p.add_argument("-c", "--coordinates", help="file to write coordinates")
    p.add_argument("-H", "--histogram-information", help="file to write histogram information")
    p.add_argument("-o", "--output", help="output file")
    p.add_argument("-w", "--weight-type", default="atan", help="weight type (atan(default),linear,none)")
    p.add_argument("--reorder-process", default=False, action="store_true")
    p.add_argument("input", help="input file")
    return p


def atan_weight_function(c, p):
    def func(b, d):
        return math.atan(c * abs(d - b) ** p)

    return func


def linear_weight_function(death_max):
    def func(b, d):
        return abs(d - b) / death_max

    return func


def main(args=None):
    args = args or argument_parser().parse_args()
    diagram = pdgm.load(args.input, args.type, args.degree, args.negate)
    check_args(args)

    histogram = tuned_histogram(diagram, args)

    if args.coordinates:
        np.savetxt(args.coordinates, histogram.centers_of_vectorize_bins())

    if args.histogram_information:
        save_histogram_information(args.histogram_information, histogram.histospec)

    if args.output:
        np.savetxt(args.output, histogram.vectorize())
    else:
        np.savetxt(sys.stdout.buffer, histogram.vectorize())


def check_args(args):
    if args.weight_type == "atan" and args.C is None:
        raise RuntimeError('weigth "atan" requires option -C')


def tuned_histogram(diagram, args):
    xy_rulers = Ruler.create_xy_rulers(args.x_range, args.xbins, args.y_range, args.ybins, diagram)
    histogram = PDHistogram(diagram, *xy_rulers)

    def apply_weight():
        if args.weight_type == "atan":
            histogram.apply_weight(atan_weight_function(args.C, args.p))
        elif args.weight_type == "linear":
            histogram.apply_weight(linear_weight_function(np.max(np.abs(diagram.deaths - diagram.births))))
        elif args.weight_type == "none":
            pass
        else:
            raise "Unknown weight type {}".format(args.weight_type)

    def apply_gaussian():
        histogram.apply_gaussian_filter(args.gaussian_sd)

    if args.reorder_process:
        apply_gaussian()
        apply_weight()
    else:
        apply_weight()
        apply_gaussian()

    return histogram


def save_histogram_information(path, histospec):
    with open(path, "w") as f:
        json.dump(histogram_info_dict(histospec), f)


def histogram_info_dict(histospec):
    yindices, xindices = histospec.indices_for_vectorization()
    return {
        "x-edges": histospec.xedges.tolist(),
        "y-edges": histospec.yedges.tolist(),
        "x-indices": xindices.tolist(),
        "y-indices": yindices.tolist(),
        "sign-flipped": histospec.sign_flipped,
    }


if __name__ == "__main__":
    main()



# pylint: disable=R0204
import os
import subprocess
import sys
import threading

import numpy as np


def paraview_programname():
    return os.environ.get("HOMCLOUD_PARAVIEW_PROGRAMNAME", "paraview")


def invoke_paraview(*args, wait=False, finalize=lambda: None):
    devnull = subprocess.DEVNULL
    redirect = {"stdin": devnull, "stdout": devnull, "stderr": devnull}
    program_name = paraview_programname()

    def invoke():
        if sys.platform.startswith("darwin"):
            subprocess.check_call(["open", "-a", program_name, "-W", "--args"] + list(args), **redirect)
        elif sys.platform.startswith("win"):
            subprocess.check_call([program_name] + list(args), **redirect)
        else:
            subprocess.check_call(["nohup", program_name] + list(args), **redirect)

        finalize()

    if wait:
        invoke()
    else:
        threading.Thread(target=invoke).start()


def deep_tolist(lst):
    """Convert ndarray to python's list recursively.

    This function is useful to convert data into json/msgpack format.

    TODO: support dict.
    """
    if isinstance(lst, np.ndarray):
        return lst.tolist()
    if isinstance(lst, list):
        return [deep_tolist(l) for l in lst]
    return lst


def load_symbols(path):
    if path is None:
        return None

    symbols = []
    with open(path) as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            symbols.append(line)
    return symbols



import argparse
import json
import sys
import numpy as np
from sklearn import mixture, cluster

from . import histo


def argument_parser():
    p = argparse.ArgumentParser(description="Output fitting information with json format")
    # p.add_argument("-n", default=2, type=int, help="Number of component")
    p.add_argument("-M", "--method", default="gmm", help="(k-means|gmm)")
    p.add_argument("-N", "--noise", metavar="SD", type=float, default=None, help="Additive gaussian noise")
    p.add_argument("-p", "--pixelvalue-range", metavar="RANGE", action=histo.RangeAction)
    p.add_argument("input", help="Input image file name")
    p.add_argument("output", nargs="?", help="Output json file name")
    return p


def fit_gmm(pixels):
    gmm = mixture.GMM(n_components=2)
    gmm.fit(pixels)
    dists = histo.WeightedNormal.from_gmm(gmm)

    boundary = None
    for k in range(int(dists[0].mean), int(dists[1].mean)):
        if dists[0].density(k) < dists[1].density(k):
            boundary = k
            break

    return {
        "boundary": boundary,
        "gaussians": [
            {"mean": dists[0].mean, "sd": dists[0].sd, "weight": dists[0].weight},
            {"mean": dists[1].mean, "sd": dists[1].sd, "weight": dists[1].weight},
        ],
    }


def fit_kmeans(pixels):
    k_means = cluster.KMeans(2)
    k_means.fit(pixels)

    return 0


def main(args):
    pixels = histo.load_pixels(args.input)

    if args.noise is not None:
        pixels = pixels + args.noise * np.random.randn(*pixels.shape)

    if args.pixelvalue_range:
        pixels = histo.filter_by_range(pixels, *args.pixelvalue_range)

    pixels = pixels.reshape((pixels.size, 1))

    if args.method == "gmm":
        dic = fit_gmm(pixels)
    elif args.method == "k-means":
        dic = fit_kmeans(pixels)

    if args.output:
        with open(args.output, "w") as f:
            json.dump(dic, f)
    else:
        json.dump(dic, sys.stdout)


if __name__ == "__main__":
    main(argument_parser().parse_args())



import numpy as np


def volume_range(volume):
    return np.min(volume, axis=0), np.max(volume, axis=0) + 1


def volrange_to_slices(range):
    return tuple(slice(b, e) for (b, e) in zip(*range))


def volume_to_bitmap(volume, range):
    min, max = range
    shape = max - min
    indices = np.ravel_multi_index(np.transpose(volume) - min[:, np.newaxis], shape)
    arr = np.zeros(shape=shape, dtype=bool)
    np.put(arr, indices, True)
    return arr



import argparse
import sys

import numpy as np

from homcloud.alpha_filtration import AlphaShape
from homcloud.version import __version__
from homcloud.license import add_argument_for_license
from homcloud.argparse_common import parse_bool, check_abolished_output
from homcloud.utils import load_symbols


def argument_parser():
    parser = argparse.ArgumentParser(description="Convert a point cloud to dipha's input (boundary matrix)")
    parser.add_argument("-V", "--version", action="version", version=__version__)
    parser.add_argument("-t", "--type", default="text", help="input file format type")
    parser.add_argument("-n", "--noise", type=float, default=0.0, help="level of additive noise")
    parser.add_argument("-d", "--dimension", type=int, default=3, help="dimension of the input data")
    parser.add_argument(
        "-w", "--weighted", action="store_true", default=False, help="use an weighted alpha filtration"
    )
    parser.add_argument("--square", action="store_true", default=True)
    parser.add_argument(
        "--no-square",
        action="store_const",
        const=False,
        dest="square",
        help="no squared output, if a birth radius is negative, the output is -sqrt(abs(r))",
    )

    parser.add_argument(
        "-P",
        "--partial-filtration",
        default=False,
        action="store_true",
        help="Compute partial filtration (relative homology)",
    )
    parser.add_argument(
        "-A", "--check-acyclicity", default=False, action="store_true", help="Check acyclicity for paritial filtration"
    )
    parser.add_argument(
        "--save-suppl-info", default=True, type=parse_bool, help="save supplementary information of PD"
    )
    parser.add_argument(
        "-M",
        "--save-boundary-map",
        default=True,
        type=parse_bool,
        help="save boundary map into output file" "(only available with phat-* algorithms, *on*/off)",
    )
    parser.add_argument(
        "--save-phtrees",
        default=False,
        type=parse_bool,
        help="save phtrees into output pdgm file" "(only available with phat-* algorithms, *on*/off)",
    )
    parser.add_argument(
        "--algorithm", default=None, help="algorithm (dipha, phat-twist(default), " "phat-chunk-parallel)"
    )
    parser.add_argument("--vertex-symbols", help="vertex symbols file")
    parser.add_argument(
        "--periodicity",
        nargs=6,
        type=float,
        default=None,
        metavar=("xmin", "xmax", "ymin", "ymax", "zmin", "zmax"),
        help="use a periodic alpha filtration",
    )
    add_argument_for_license(parser)
    parser.add_argument("input", metavar="INPUT", help="input file name")
    parser.add_argument("output", metavar="OUTPUT", help="output file name")
    return parser


def noise_array(level, dim, weighted, partial, num_points):
    noise = np.random.uniform(-level, level, (num_points, dim))
    if weighted:
        noise = np.hstack([noise, np.zeros((num_points, 1))])
    if partial:
        noise = np.hstack([noise, np.zeros((num_points, 1))])
    return noise


def parse_periodicity(p):
    return ((p[0], p[1]), (p[2], p[3]), (p[4], p[5])) if p else None


def main(args=None):
    args = args or argument_parser().parse_args()
    if args.dimension == 2 and args.weighted:
        raise RuntimeError("Weighted 2D alpha filtration is not supported now")

    check_abolished_output(args)

    points = np.loadtxt(args.input)

    if points.shape[1] != width(args):
        raise RuntimeError("Input format error")

    if args.noise > 0.0:
        points += noise_array(args.noise, args.dimension, args.weighted, args.partial_filtration, points.shape[0])

    alpha_shape = AlphaShape.build(
        points, args.dimension, args.weighted, args.partial_filtration, parse_periodicity(args.periodicity)
    )

    if args.check_acyclicity:
        alpha_shape.check_subsets_acyclicity()

    if args.partial_filtration:
        alpha_shape.become_partial_shape()

    filtration = alpha_shape.create_filtration(
        args.square, load_symbols(args.vertex_symbols), args.save_boundary_map, args.save_phtrees
    )
    filtration.compute_pdgm_and_save(args.output, args.algorithm, args.save_suppl_info)


def width(args):
    return args.dimension + (1 if args.weighted else 0) + (1 if args.partial_filtration else 0)


if __name__ == "__main__":
    sys.exit(main())
